/*  >>> this is file IOTOOLS.C
        compile it with TURBO-C in large model
============================================================================
   LSETUP v1.3 (C) 1995..1996 Hans Lermen (lermen@elserv.ffm.fgan.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

------------------------------------------------------------------------------

  NOTE: disk write code is a modified RAWRITE.EXE 1.2
        originally written by Mark Becker,
        so, this part is (C) Mark Becker.
        He does not mention any copyright restrictions in the
        source, and as this source is distributed round over the
        world on allmost any Linux FTP-site, I took it and used
        it for my purpose.
        However, I don't know if Mark Becker did put it under GPL.


============================================================================
*/

#if 0
  #define TESTING
#endif

#ifndef TESTING
#include "_malloc_.h" /* malloc, realloc intercept, must be placed TOP */
#endif

#include <stdlib.h>
#include <alloc.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <setjmp.h>
#include <dir.h>

#include "lvideo.h"
#include "iotools.h"

#ifndef TESTING
#include "parserd.h"
#include "setup.h"
#endif

#include <bios.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>

#define space64k 0x10000

static void *DMAbuffer=0;

void fxpand(char *path, char *fname, char *defaultext)
{
  char drive[MAXDRIVE],dir[MAXDIR],name[MAXFILE],ext[MAXEXT];
  int flag=fnsplit(fname,drive,dir,name,ext);
  if (!(flag & DRIVE)) {
    drive[0]='A' + getdisk();
    drive[1]=':';
    drive[2]=0;
  }
  else drive[0]=toupper(drive[0]);
  if (!(flag & DIRECTORY)) {
    getcurdir(1+drive[0]-'A',dir+1);
    dir[0]='\\';
    if (dir[1]) strcat(dir,"\\");
  }
  if (defaultext && defaultext[0] && !(flag & EXTENSION)) {
    ext[0]='.';
    if (defaultext[0] == '.') defaultext++;
    strcpy(ext+1,defaultext);
  }
  fnmerge(path,drive,dir,name,ext);
}

struct device_driver_header *get_driver(char *name)
{
  struct REGPACK r;
  struct device_driver_header *d;
  char n[9];
  int i;

  r.r_ax=0x122c;
  r.r_bx=0;
  intr(0x2f,&r);
  if ((!r.r_bx) || (r.r_ax==0x122c)) return 0;
  d=MK_FP(r.r_bx,r.r_ax);
  i=strlen(name);
  if (i>8) i=8;
  memset(n,' ',8);
  memcpy(n,name,i);
  i=50; /* in case we are scanning garbage, we limit the loop */
  while ((FP_OFF(d) !=0xffff) && i--) {
    if (!strncmp(d->name,n,8)) return d;
    d=d->next;
  }
  return 0;
}



#if 0
static void *malloc_DMAbuffer(unsigned short size)
{
  unsigned long l, l_;
  void **p;
  char *p_;
  p=farmalloc(size+space64k+sizeof(void *));
  l=((long)FP_SEG(p) << 4) + FP_OFF(p);
  l_=(l+space64k-1) & (~(space64k-1));
  if (l_ < (l+sizeof(void *)) ) l_+=sizeof(void *);
  p_= (void *) (((l_ & ~15)<<12)+(l_&15));
  *((void **)(p_-sizeof(void *)))=(void *)p; /*save the original pointer */
  return p_;
}
static void free_DMAbuffer(void **pp)
{
  void *p=*pp;
  if (!p) return;
  p= *((void **)(((char *)p)-sizeof(void *)));
  farfree(p);
  *pp=0;
}
#else

#define MAXDMASIZE (18*2*512)
static char *__buf=0;

static void *malloc_DMAbuffer(unsigned short size)
{
  unsigned long l, boundary;
  void **p;
  char *p_;
  if (!__buf) __buf=malloc(MAXDMASIZE);
  p=(void **)__buf;
  l=((long)FP_SEG(p) << 4) + FP_OFF(p);
  boundary= (l+space64k-1) & (~(space64k-1));
  if ((l+size) > boundary) l=boundary;
  p_= (void *) (((l & ~15)<<12)+(l&15));
  return p_;
}
static void free_DMAbuffer(void **pp)
{
}

#endif


static done;

static cancel()
{
  if (done) return done;
  done=check_for_keystroke("\x1b");
}


unsigned char get_cmos_byte(int index)
{
  enum {CMOS_INDEX=0x70,CMOS_DATA};
  outportb(CMOS_INDEX,index);
  return inportb(CMOS_DATA);
}

#if 0
int CMOS_harddisk_type(int drive)
{
  enum {CMOS_HARDDISKS=0x12,CMOS_HARDDISK0=0x19,CMOS_HARDDISK1};
  int h,r;
  h = get_cmos_byte(CMOS_HARDDISKS);
  if (drive) {
    h &=15;
    drive=1;
  }
  else h >>=4;
  if (h<15) return h;
  r = get_cmos_byte(CMOS_HARDDISK0+drive);
  return r;
}
#endif

/* Gets installed RAM in units of 1 KBytes */
unsigned get_installed_mem_from_CMOS()
{
  enum {CMOS_INSTALLED_RAM=0x17};
  unsigned ram;
  ram =  get_cmos_byte(CMOS_INSTALLED_RAM);
  ram = ((long)get_cmos_byte(CMOS_INSTALLED_RAM+1)) << 8;
  return ram +0x400;
}

static void *oldint1e=0;


void free_dpt()
{
  if (oldint1e) {
    setvect(0x1e,(void interrupt (*) ())oldint1e);
    oldint1e=0;
  }
}

void set_dpt()
{
  static unsigned char our_dpt[] ={
    0xdf,0x02,0x25,0x02,0x12,0x1b,0xff,0x54,0xf6,0x0f,0x08
  };
  if (oldint1e) free_dpt();
  oldint1e=(void *)getvect(0x1e);
  setvect(0x1e,(void interrupt (*) ())&our_dpt);
}

/*=========================================================================*/
/* NOTE: the following disk write code is a modified RAWRITE.EXE
         originally written by Mark Becker, here is his header

 rawrite.c	Write a binary image to a 360K diskette.
		By Mark Becker

 Usage:
	MS-DOS prompt> RAWRITE

	And follow the prompts.

History
-------

  1.0	-	Initial release
  1.1	-	Beta test (fixing bugs)				4/5/91
  		Some BIOS's don't like full-track writes.
  1.101	-	Last beta release.				4/8/91
  		Fixed BIOS full-track write by only
		writing 3 sectors at a time.
  1.2	-	Final code and documentation clean-ups.		4/9/91
*/


#define FALSE	0
#define TRUE	(!FALSE)

#define SECTORSIZE	512

#define	RESET	0
#define	LAST	1
#define	READ	2
#define	WRITE	3
#define	VERIFY	4
#define	FORMAT	5

#define SECPERCALL -1 /* or 3 or 1 */
#define MAXRETRY 10

#if 0 /* original */


void msg(char (*s))
{
	printf("%s\n", s);
	_exit(1);
}


/*
 Identify the error code with a real error message.
*/
static void Error(int (status))
{
  switch (status) {
    case 0x00:	msg("Operation Successful");				break;
    case 0x01:	msg("Bad command");					break;
    case 0x02:	msg("Address mark not found");				break;
    case 0x03:	msg("Attempt to write on write-protected disk");	break;
    case 0x04:	msg("Sector not found");				break;
    case 0x05:	msg("Reset failed (hard disk)");			break;
    case 0x06:	msg("Disk changed since last operation");		break;
    case 0x07:	msg("Drive parameter activity failed");			break;
    case 0x08:	msg("DMA overrun");					break;
    case 0x09:	msg("Attempt to DMA across 64K boundary");		break;
    case 0x0A:	msg("Bad sector detected");				break;
    case 0x0B:	msg("Bad track detected");				break;
    case 0x0C:	msg("Unsupported track");				break;
    case 0x10:	msg("Bad CRC/ECC on disk read");			break;
    case 0x11:	msg("CRC/ECC corrected data error");			break;
    case 0x20:	msg("Controller has failed");				break;
    case 0x40:	msg("Seek operation failed");				break;
    case 0x80:	msg("Attachment failed to respond");			break;
    case 0xAA:	msg("Drive not ready (hard disk only");			break;
    case 0xBB:	msg("Undefined error occurred (hard disk only)");	break;
    case 0xCC:	msg("Write fault occurred");				break;
    case 0xE0:	msg("Status error");					break;
    case 0xFF:	msg("Sense operation failed");				break;
  }
  _exit(1);
}

#else

#endif


/*
 Identify what kind of diskette is installed in the specified drive.
 Return the number of sectors per track assumed as follows:
 9	-	360 K and 720 K 5.25".
15	-	1.2 M HD	5.25".
18	-	1.44 M		3.5".
*/
static int nsects(int (drive), char *nsects_ft)
{
  static	int	nsect[] = {18, 15, 9};

  char	*buffer;
  int	i, status;
/*
 Read sector 1, head 0, track 0 to get the BIOS running.
*/
  DMAbuffer= malloc_DMAbuffer(512);
  buffer= DMAbuffer;
  biosdisk(RESET, drive, 0, 0, 0, 0, buffer);
  status = biosdisk(READ, drive, 0, 10, 1, 1, buffer);
  if (status == 0x06)			/* Door signal change?	*/
  status = biosdisk(READ, drive, 0, 0, 1, 1, buffer);

  for (i=0; i < sizeof(nsect)/sizeof(int); ++i) {
    biosdisk(RESET, drive, 0, 0, 0, 0, buffer);
    status = biosdisk(READ, drive, 0, 0, nsect[i], 1, buffer);
    if (status == 0x06)
      status = biosdisk(READ, drive, 0, 0, nsect[i], 1, buffer);
      if (status == 0x00) break;
    }
    if (i == sizeof(nsect)/sizeof(int)) {
      cprintf(nsects_ft);
      return 0;
    }
    return(nsect[i]);
}

/*                    +------name of input file
                      |       +-- drive letter of floppy (e.g. 'A')
                      |       |            +-- format for printf like "Can't figure out how many sectors/track for this diskette."
                      |       |            |                                */
int disk_rawrite(char	*fname, char drive, char *nsects_ft, char *errio_ft, char *THS_ft, char *done_ft)
/*                                                                |           |            |
    format for printf like "Disketten IO-Fehler, Status %2x\n" ---+           |            |
    format for printf like "Track: %02d  Head: %2d Sector: %2d\r" ------------+            |
    format for printf like "\nSchreiben auf Diskette beendet\n" ---------------------------+ */
{
  char	*buffer, *pbuf;
  int	 count, fdin, head, track, status, spt, buflength, ns;
  int ret,retry,secpercall;
  if ((fdin = open(fname, O_RDONLY | O_BINARY)) <= 0) {
     return(0);
  }

  done = 0;
  drive = toupper(drive) -'A';


/*
 * Determine number of sectors per track and allocate buffers.
 */
  set_dpt();
  spt = nsects(drive,nsects_ft);
  free_dpt();
  cprintf("\nwriting %d sectors per track\n\n",spt);
  buflength = spt * SECTORSIZE;
  free_DMAbuffer(&DMAbuffer);
  if (!spt) return 0;
  DMAbuffer= malloc_DMAbuffer(buflength);
  buffer = (char *)DMAbuffer;

  set_dpt();
  biosdisk(RESET, drive, 0, 0, 0, 0, buffer);

  if (SECPERCALL <0) secpercall=spt;
  else secpercall=SECPERCALL;
/*
 * Start writing data to diskette until there is no more data to write.
 */
   head = track = 0;
   while (((count = read(fdin, buffer, buflength)) > 0) && (!cancel())) {
     pbuf = buffer;
     for (ns = 1; count > 0 && !cancel(); ns+=secpercall) {
       cprintf(THS_ft, track, head, ns);
       retry=MAXRETRY;
       do {
         status = biosdisk(WRITE, drive, head, track, ns, secpercall, pbuf);
       } while ( (status == 4) && (--retry >0) );

       if (status != 0) {
         cprintf("\n");
         cprintf(errio_ft,status);
         free_DMAbuffer(&DMAbuffer);
         close(fdin);
         free_DMAbuffer(&DMAbuffer);
         free_dpt();
         return 0;
       }

       count -= (secpercall*SECTORSIZE);
       pbuf  += (secpercall*SECTORSIZE);
     }
     if ((head = (head + 1) & 1) == 0) ++track;
   }
   if (eof(fdin)) {
     cprintf(done_ft);
     biosdisk(2, drive, 0, 0, 1, 1, buffer);		/* Retract head	*/
     ret=1;
   }
   else ret=0;
   free_DMAbuffer(&DMAbuffer);
   free_dpt();
   close(fdin);
   return ret;
}

/* below only for testing *************************** */

#ifdef TESTING

char err_mesg[256];

void err_exit(char* string, ...)
{
  va_list vars;
  va_start(vars, string);
  if (vsprintf(err_mesg, string, vars) == EOF) strcpy(err_mesg, string);
  va_end(vars);
  if (err_mesg[0]) {
    /* we must use the BIOS to print the string, because we
       may come here due to heap overflow */
    struct REGPACK r;
    unsigned char *s=(void *)err_mesg;
    while (*s) {
      if (*s == '\n') {
        r.r_ax=0x0e0d;
        r.r_bx=0;
        intr(0x10,&r);
      }
      r.r_ax=0x0e00 + *s;
      r.r_bx=0;
      intr(0x10,&r);
      s++;
    };
  }
  exit(1);
}



main(int argc, char **argv)
{
#if 0
  printf("installed ram = 0x%08x (%u)\n",get_installed_mem_from_CMOS(),get_installed_mem_from_CMOS());
  exit (0);
#endif
  initvideo();
  cprintf("started\n");
  cprintf("\nreturn %d\n",
    disk_rawrite(argv[1],argv[2][0],
      "Can't figure out how many sectors/track for this diskette.",
      "Disketten IO-Fehler, Status %2x\n",
      "Track: %02d  Head: %2d Sector: %2d\r",
      "\nSchreiben auf Diskette beendet\n"
    )
  );
}
#endif