%{
/*  >>> this is file lexer.l
        use the GNU flex to generate lexer.c
        lexer.c should compile with GCC and with TURBO-C in large model
============================================================================
   LSETUP v1.0 (C) 1995 Hans Lermen (lermen@elserv.ffm.fgan.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


============================================================================
*/
#ifndef __linux__
  #include "_malloc_.h" /* malloc, realloc intercept, must be placed TOP */
#else
  #define strdup our_strdup
#endif
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
 #include <setjmp.h>
 #include <sys/stat.h>                    /* structure stat       */
#ifdef __linux__
 #include <unistd.h>                      /* prototype for stat() */
#endif
#include <math.h>

#include "parser.h"
#include "parserd.h"
int line_count;

static int __active__=1, __next_ITEM_for_me__=0;
#define __active_max 15
static int __active_stack[__active_max+1]={1,0};
static int __active_i=1;

static void push__active();
static void pop__active();
static void enter_includefile(char * fname);

extern char* strdup(const char *); /* Not defined in string.h :-( */
extern void yyerror(char* string, ...);

char *yy_vbuffer=0;

#define YY_INPUT(buf,result,max_size) \
  if (!yyin) { \
    if (yy_vbuffer && yy_vbuffer[0]) { \
      buf[(max_size)-1]=0; \
      strncpy(buf,yy_vbuffer,max_size); \
      if (buf[(max_size)-1]) { \
        yy_vbuffer+=max_size; \
        result=max_size; \
      } \
      else { \
        result=strlen(buf); \
        yy_vbuffer=0; \
      } \
    } \
    else result=0; \
  } \
  else { \
    if ( yy_current_buffer->yy_is_interactive ) { \
      int c = getc( yyin ); \
      result = c == EOF ? 0 : 1; \
      buf[0] = (char) c; \
    } \
    else { \
      if ( ((result = fread( buf, 1, max_size, yyin )) == 0) && ferror( yyin ) ) \
                  YY_FATAL_ERROR( "input in flex scanner failed" ); \
    } \
  }



#undef YY_DECL
#define YY_DECL int yylex YY_PROTO(( YYSTYPE* yylval ))

%}

%%
     
	/* special characters
 ,   =   (   )   {   }   :   +   -   *   /   %   |   &   ^   ~  */
","|"="|"("|")"|"{"|"}"|":"|"+"|"-"|"*"|"/"|"%"|"|"|"&"|"^"|"~" if (__active__) return(yytext[0]);

	/* keywords */

menu			if (__active__) return(MENU);
image			if (__active__) return(IMAGE);
action			if (__active__) return(ACTION);
function		if (__active__) return(FUNCTION);
show			if (__active__) return(SHOW);
waitkey			if (__active__) return(WAITKEY);
printxy			if (__active__) return(PRINTXY);
string			if (__active__) return(TSTRING);
void			if (__active__) return(TVOID);
integer			if (__active__) return(TINTEGER);
start			if (__active__) return(START);
call			if (__active__) return(CALL);
goto			if (__active__) return(GOTO);
iftrue			if (__active__) return(IFTRUE);
iffalse			if (__active__) return(IFFALSE);
label			if (__active__) return(LABEL);
set			if (__active__) return(SET);
colors			if (__active__) return(COLORS);
iferror			if (__active__) return(IFERROR);
ifnoterror		if (__active__) return(IFNOTERROR);
"#ifdef"		{ push__active(); if (__active__) __next_ITEM_for_me__=1; }
"#ifndef"		{ push__active(); if (__active__) __next_ITEM_for_me__=-1; }
"#else"			{ 
			  if (__active__) __active__=0;
			  else if (__active_stack[__active_i-1] ) __active__=1;
			}
"#endif"		pop__active();
"#include"		if (__active__) __next_ITEM_for_me__=2;


	/* operators */
">>"			if (__active__) return(SHR_OP);
"<<"			if (__active__) return(SHL_OP);

	/* integers */

(([0-9]+)|(0x[0-9a-fA-F]+))	if (__active__) { yylval->i_value = strtol(yytext, NULL, 0);
			  return(UINTEGER); }


	/* strings */

\"[^\"]*\"		if (__active__) {
			  yytext[strlen(yytext)-1] = '\0';
			  if (__next_ITEM_for_me__ == 2) {
                            __next_ITEM_for_me__=0;
			    enter_includefile(&yytext[1]);
			  }
			  else {
			    yylval->s_value = strdup(&yytext[1]);
			    return(STRING);
			  }
			}
\'[^\']*\'		if (__active__) { yytext[strlen(yytext)-1] = '\0';
			  yylval->s_value = strdup(&yytext[1]);
			  return(STRING); }
[a-zA-Z_]+[0-9a-zA-Z_]*	{
			  if (__next_ITEM_for_me__) {
			    if (get_object_by_name(yytext,OBJ_VARSTRING)) {
			      __active__= (__next_ITEM_for_me__ > 0);
			    }
			    else __active__= (__next_ITEM_for_me__ < 0);
			    __next_ITEM_for_me__=0;
			  }
			  else if (__active__) {
			    yylval->s_value = strdup(yytext);
			    if (FIRST_PASS) return(IDENT);
			    else return(IDENT2);
			  }
			}

	/* comments & whitespace */

[\;\!][^\n]*\n		line_count++;   /* comments to (and including) EOLN */
[ \t\r]*		;   		/* ignore all white space */
\n			line_count++;	/* keep track of lines seen */
\032			;		/* ignore DOSish EOF (^Z) */
.			fprintf(stderr, "line %d: discarding char '%c' = 0x%02x\n", line_count, yytext[0], yytext[0] & 0xff);


%%

#define MAX_INCLUDE_DEPTH 10
static YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH] ={0};
int include_stack_ptr = 0;
char * include_fnames[MAX_INCLUDE_DEPTH] = {0};
static int include_lines[MAX_INCLUDE_DEPTH] = {0};

static void enter_includefile(char * fname)
{
  FILE * new_yyin;
  char fname_[256];
  if ( include_stack_ptr >= MAX_INCLUDE_DEPTH ) {
    yyerror("Includes nested too deeply" );
    return;
  }
  /* we want to have the include files in the same directory as
   * the main menu-file, if we have no leading '\\' or '/'
   */
  if ( (fname[0] != '/') && (fname[0] != '\\') 
     && (fname[1] != ':') ) {
    int i;
    strcpy(fname_,include_fnames[0]);
    i=strlen(fname_);
    while (i && (fname_[i] != '\\') && (fname_[i] != '/') && (fname_[i] != ':')) i--;
    if (i) {
      i++;
      strcpy(fname_+i,fname);
      fname=fname_;
    }
  }
  new_yyin = fopen( fname, "r" );
  if ( ! new_yyin ) {
    yyerror("cannot open includefile %s", fname);
    return;
  }
  include_lines[include_stack_ptr] = line_count;
  include_stack[include_stack_ptr] = YY_CURRENT_BUFFER;
  include_stack_ptr++;
  include_fnames[include_stack_ptr] = strdup(fname);
  line_count = 1;
  yyin = new_yyin;
  include_stack[include_stack_ptr] = yy_create_buffer( yyin, YY_BUF_SIZE );
  yy_switch_to_buffer(include_stack[include_stack_ptr]);
}


#ifdef yywrap
  error "yywrap defined elsewere, need our own one"
#endif

int yywrap(void)    /* do this to avoid needing to do -lfl */
{
  if (include_stack_ptr <= 0 ) return(1);
  yy_switch_to_buffer(include_stack[include_stack_ptr-1] );
  fclose(include_stack[include_stack_ptr]->yy_input_file);
  yy_delete_buffer(include_stack[include_stack_ptr]);
#if 0
  printf(">>>>>>>>>%s<\n",include_fnames[include_stack_ptr]); 
#endif
  free(include_fnames[include_stack_ptr]);
  include_stack_ptr--;
  line_count=include_lines[include_stack_ptr];
  return(0);
}


/* we do this to get some GCC warnings out,
 * without generally switching off all warnings of this kind.
 */  
void *yy_avoid_warning_stuff_list[]={
  yyunput,
  yy_push_state,
  yy_pop_state,
  yy_top_state
};

static void push__active()
{
  __active_stack[__active_i++]=__active__;
  if (__active_i > __active_max) {
    __active_i = __active_max;
    yyerror("Lexer block stack overflow, unmatching #if..#else..#endif");
  }
}

static void pop__active()
{
  if (__active_i <=0) {
    yyerror("Lexer block stack underflow, unmatching #if..#else..#endif");
    return;
  }
  __active__=__active_stack[--__active_i];
}

