/* This is file LVIDEO.C
   Version 2.1
============================================================================

   (C) 1992..1996 Hans Lermen, D5480 Remagen, GERMANY
                  (lermen@elserv.ffm.fgan.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


============================================================================

   Direct dual-monitor video package,
   without (or a minimum) of PC-BIOS-calls.
   
   If __bios_switch__ is not defined, then NONE bios calls is done!
   All is done by directly accessing the only well known PC hardware.
   This is faster and (more important) avoids BIOS incompatibilities
   and last not least can be run by other CPU then the ix86 family
   (i.e. i860, 68K, and others on an ISA/EISA architecture )

   This Package has been successful compiled by the following C-Compiler:

   TURBO-C  (Borland)
   DJGCC    (Free Software Foundation, D.J.Delorie's DOS-portation)
   HIGH-C   (MetaWare, i860 Compiler)

*/

/* NOTE: this is an quit, but not optimal fix of an error !!!!!
 *       have to fix this later
 */

#include "_malloc_.h" /* malloc, realloc intercept, must be placed TOP */

#define LVIDEO_itself

#ifdef __TURBOC__
  #include "lvideo.h"
  #include <stdlib.h>
  #include <stdarg.h>
  #include <string.h>
  #ifdef __bios_switch__
    #include <dos.h>
  #endif
  #define bios_data 0x00400000
  #define mono_screen_buffer  0xb0000000
  #define color_screen_buffer 0xb8000000
#elif defined(__GNUC__)
    #include "lvideo.h"
    #include <std.h>
    #include <pc.h>
    #include <no_varg.h>  /* LERMEN */
    #include <stdarg.h>
    #define bios_data 0xE0000400
    #define mono_screen_buffer  0xE00b0000
    #define color_screen_buffer 0xE00b8000
#else
  #include "lvideo.h"
  #include <stdlib.h>
  #include <stdarg.h>
  #include <string.h>
  #ifdef __i860paged__
    #define bios_data 0xF8000400
    #define mono_screen_buffer  0xF80b0000
    #define color_screen_buffer 0xF80b8000
  #else
    #define bios_data 0x00000400
    #define mono_screen_buffer  0x000b0000
    #define color_screen_buffer 0x000b8000
  #endif
#endif

enum { mono_crt_port=0x3b4, hgc_config_port=0x3bf, color_crt_port=0x3d4};
enum { cursor_start_reg=10,cursor_end_reg,
       cursor_loc_high=14,cursor_loc_low};


int bios_current_screen=this_screen;
screentyp * masters[3] = {0,0,0};

char *cprintf_buffer=0;

#ifdef __TURBOC__
  #if 1
    #define __bios_switch__
    #include <dos.h>
  #endif
#endif


#ifdef __bios_switch__
  #ifdef __TURBOC__
    int videoint(int ah,int al){
      struct REGPACK reg;
      reg.r_ax=(ah<<8) + al;
      intr(0x10,&reg);
    }
  #elif defined(__GNUC__)
    int videoint(int ah,int al){
      /* avoid it !!! */
    }
  #else
    int videoint(int ah,int al){
      /* avoid it !!! */
    }
  #endif
#endif

static int read_crt_reg(int baseport,int index)
{
  outportb(baseport,index);
  return inportb(baseport+1);
}

static void write_crt_reg(int baseport,int index,int value)
{
  outportb(baseport,index);
  outportb(baseport+1,value);
}

static save_bios_screen_state(screentyp *s)
{
  s->equipflag=*((word *)(bios_data+0x10));
  memcpy(&s->current_mode,(byte *)(bios_data+0x49),0x66-0x49+1);
  memcpy(&s->egavga_params[0],(byte *)(bios_data+0x84),7);
  memcpy(&s->egapointer[0],(byte *)(bios_data+0xa8),4);
}

static restore_bios_screen_state(screentyp *s)
{
  *((word *)(bios_data+0x10)) = s->equipflag;
  memcpy((byte *)(bios_data+0x49),&s->current_mode,0x66-0x49+1);
  memcpy((byte *)(bios_data+0x84),&s->egavga_params,7);
  memcpy((byte *)(bios_data+0xa8),&s->egapointer[0],4);
}



static int is_writable_ram(dword address)
{
  word save,value;
  save=*((word *)(address));
  *((word *)(address))+=0x1234;
  value=++(*((word *)(address)));
  *((word *)(address))=save;
  return ((value & 0xffff) == ((save + 0x1234 +1) & 0xffff)) ;
}


static int read_hardware_cursor(int baseport)
{
  return read_crt_reg(baseport,cursor_loc_low)+(read_crt_reg(baseport,cursor_loc_high)<<8);
}

static void write_hardware_cursor(int baseport,int value)
{
  write_crt_reg(baseport,cursor_loc_low,value);
  write_crt_reg(baseport,cursor_loc_high,value>>8);
}

int check_for_egavga()
{
  screentyp s;
  save_bios_screen_state(&s);
#ifdef USE_DUALMON
  if ((s.current_mode & 0x7f) <= 3) return this_screen;
  if (   (*((word *)(color_screen_buffer+0x8000+0x1e)) == 0x4249 )
      && (*((byte *)(color_screen_buffer+0x8000+0x20)) == 0x4D ) ) {
    /* ega/vga-bios  present */
    if ((s.equipflag & 0x30) == 0x0) return this_screen;
    else return other_screen;
  }
#else
  if ((s.current_mode & 0x7f) != 7) return this_screen;
#endif
  return no_screen;
}


int check_for_graphics_mode()
{
  if (color_screen) {
    if (!read_crt_reg(color_crt_port,cursor_end_reg)) return color_screen;
    else return no_screen;
  }
}


int check_for_herkules()
{
  screentyp s;
  save_bios_screen_state(&s);
#ifdef USE_DUALMON
  if (    ((s.current_mode & 0x7f) == 7)
       && ((s.equipflag & 0x30) == 0x30) ) return this_screen;
  outportb(hgc_config_port,0); /* put HGC into DIAG */
  if (is_writable_ram(mono_screen_buffer)) {
    if (is_writable_ram(mono_screen_buffer+0x1000)){
      /* check for 4Kb-modulo addressing */
      word i0 = *((word *)(mono_screen_buffer));
      word i1 = *((word *)(mono_screen_buffer+0x1000));
      int r;
      *((word *)(mono_screen_buffer)) = *((word *)(mono_screen_buffer+0x1000)) = 0;
      ++(*((word *)(mono_screen_buffer)));
      if (*((word *)(mono_screen_buffer+0x1000)) == 1) r=other_screen;
      else r=no_screen;
      *((word *)(mono_screen_buffer))=i0;
      *((word *)(mono_screen_buffer+0x1000))=i1;
      if (r==no_screen){
        int c;
        write_hardware_cursor(mono_crt_port,(c=read_hardware_cursor(mono_crt_port))+1);
        if (read_hardware_cursor(mono_crt_port)==(c+1)){
          r=other_screen;
          write_hardware_cursor(mono_crt_port,c);
        }
      }
      return r;
    }
    else {
      /* RAM behind page 0 is mapped out, (some clones are doing so) */
      outportb(hgc_config_port,1); /* put HGC into HALF */
      if (is_writable_ram(mono_screen_buffer+0x1000)) {
        outportb(hgc_config_port,0); /* put HGC into DIAG */
        return other_screen;
      }
      outportb(hgc_config_port,0); /* put HGC into DIAG */
    }
  }
#else
  if ((s.current_mode & 0x7f) == 7) return this_screen;
#endif
  return no_screen;
}



void update_cursor(windowtyp *w)
{
  int offs;
  screentyp *s;
  s = w->screen;
  if (!w->with_cursor) {
    write_hardware_cursor(s->crt_index, 0xFFFF /* 62*s->w.rowsize */);
    return;
  }
  s->cursors[0] = w->x1 + w->x;
  s->cursors[1] = w->y1 + w->y;
  write_hardware_cursor(s->crt_index,(s->cursors[0]+ (s->cursors[1] * s->w.rowsize)));
  if (s->w.what_screen==bios_current_screen) {
    bios_xpos = s->cursors[0];
    bios_ypos = s->cursors[1];
  }
}

void update_xy(windowtyp *w)
{
  int offs,cursor,cx,cy,x,y;
  screentyp *s;
  if (!w->with_cursor) return;
  s = w->screen;

  cursor=read_hardware_cursor(s->crt_index);
  cx=cursor % s->w.rowsize;
  cy=cursor / s->w.rowsize;
  x= cx - w->x1;
  y= cy - w->y1;
  if ( (x<0) || (y<0) || (x>=w->wrowsize) || (y>=w->wcolsize) ) return;
  s->cursors[0]=cx;
  s->cursors[1]=cy;
  w->x=x;
  w->y=y;
}

void set_global_attributes(int screen, int normattrib,
           int inversattrib, int highlightattrib, int frameattrib)
{
  screentyp *s= masters[screen];
  if (!s) return;
  s->w.normattrib=normattrib;
  s->w.inversattrib=inversattrib;
  s->w.highlightattrib=highlightattrib;
  s->w.currattrib=s->w.normattrib;
  s->w.frameattrib=frameattrib;
}

static init_current_screen()
{
  screentyp *s;
  int i,cursor;

  if (masters[current_screen]) return;

  s=(screentyp *) malloc(sizeof(screentyp));
  masters[current_screen]=s;
  #ifdef __bios_switch__
    save_bios_screen_state(s);
  #else
    if (current_screen==bios_current_screen) save_bios_screen_state(s);
    else {
      if (current_screen==color_screen) {
        *((word *)(&s->crtbase_low))=color_crt_port;
        s->current_mode=3;
        s->equipflag=*((word *)(bios_data+0x10)) & ~0x30;
      }
      else {
        *((word *)(&s->crtbase_low))=mono_crt_port;
        s->current_mode=7;
        s->equipflag=*((word *)(bios_data+0x10)) | 0x30;
      }
      s->numcolumns=80;
      s->egavga_params[0]=24;
    }
  #endif
  s->w.what_screen=current_screen;
  s->w.screen=s;
  s->w.framewindow=0;
  s->w.saveptr=0;
  s->w.with_cursor=1;
  s->w.full_screen=1;
  if (current_screen==color_screen) s->crt_index=color_crt_port;
  else s->crt_index=mono_crt_port;
  /* s->crt_index=*((word *)(&s->crtbase_low)); */
  s->crt_data=s->crt_index+1;
  s->w.wrowsize=s->numcolumns;
  s->w.rowsize=s->numcolumns;
  cursor=read_hardware_cursor(s->crt_index);
  s->cursors[0]=cursor % s->w.rowsize;
  s->cursors[1]=cursor / s->w.rowsize;
  s->w.x=s->cursors[0];
  s->w.y=s->cursors[1];
  s->w.x1 = s->w.y1 = 0;
  s->w.x2=s->w.rowsize-1;
  if (current_screen==color_screen) {
    s->regenbuf=(word *)color_screen_buffer;
    s->w.y2=s->egavga_params[0];
  }
  else {
    s->regenbuf=(word *)mono_screen_buffer;
    s->w.y2=24;
  }
  s->w.regenbuf=s->regenbuf;
  s->w.wcolsize=s->w.y2+1;
  i=s->w.x+(s->w.y*s->w.rowsize);
  if (i) i--;
  s->w.normattrib= *(s->w.regenbuf+i) & 0xff00;
  s->w.inversattrib= s->w.normattrib ^ 0x7f00;
  s->w.highlightattrib= s->w.normattrib | 0x0800;
  s->w.frameattrib=s->w.normattrib;
  s->w.currattrib=s->w.normattrib;
}

void goto_screen(int what_screen)
{
  if ((what_screen==current_screen) || (!dual_monitor))return;
  current_screen=what_screen;
  current_window=&(masters[current_screen]->w);
}


void goto_window(windowtyp *w)
{
  if (!w) return;
  goto_screen(w->what_screen);
  current_window=w;
  update_cursor(w);
}


void free_window(windowtyp *w)
{
  windowtyp *w_;
  if (w->framewindow) {
    w_=w->framewindow;
    if (w_->saveptr) free(w_->saveptr);
    free(w_);
  }
  if (w->saveptr) free(w->saveptr);
  free(w);
}

windowtyp *createwindow(int screen, int withcursor, int x1, int y1, int x2, int y2)
{
  windowtyp *w;
  if (!masters[screen]) return 0;
  w=malloc(sizeof(windowtyp));
  memcpy(w,&masters[screen]->w,sizeof(windowtyp));
  w->x=w->y=0;
  w->x1=x1;
  w->y1=y1;
  w->x2=x2;
  w->y2=y2;
  w->with_cursor=withcursor;
  w->full_screen=0;
  w->wrowsize=x2-x1+1;
  w->wcolsize=y2-y1+1;
  w->regenbuf=(w->regenbuf + (x1) + (y1*w->rowsize));
  w->framewindow=0;
  w->saveptr=0;
  return w;
}


windowtyp *createframewindow(int screen, int withcursor, int x1, int y1, int x2, int y2)
{
  windowtyp *w;
  if (!masters[screen]) return 0;
  w=createwindow(screen,withcursor,x1+1,y1+1,x2-1,y2-1);
  w->framewindow=(void *)createwindow(screen,withcursor,x1,y1,x2,y2);
  return w;
}


void initvideo()
{
  if (cprintf_buffer) return;
  cprintf_buffer=malloc(0x8000);
  cmenu_choice_preset=-1;
#ifdef USE_DUALMON
  color_screen= check_for_egavga();
  mono_screen=  check_for_herkules();
  dual_monitor= (color_screen && mono_screen);
  screen_in_graphics_mode=check_for_graphics_mode();
#else
  #if 1
    color_screen= check_for_egavga();
    mono_screen=  check_for_herkules();
  #else
    color_screen= 1;
    mono_screen= 0;
  #endif
  dual_monitor= 0;
  if (check_for_graphics_mode()) videoint(0,0x03);
  screen_in_graphics_mode=0;
#endif
  current_screen=this_screen;
  bios_current_screen=this_screen;
  init_current_screen();
  current_window=&(masters[current_screen]->w);
  this_main_window=current_window;
  if (dual_monitor) {
    #ifdef __bios_switch__
      int cursor;
      screentyp s;
      byte *save;
      current_screen=other_screen;
      save_bios_screen_state(&s);
      if (other_screen==color_screen) {
        cursor=read_hardware_cursor(color_crt_port);
        *((word *)(bios_data+0x10))&=~0x30;
        videoint(0,0x83);
        write_hardware_cursor(color_crt_port,cursor);
      }
      else {
        cursor=read_hardware_cursor(mono_crt_port);
        save=malloc(4000);
        memcpy(save,(byte *)(mono_screen_buffer),4000);
        *((word *)(bios_data+0x10))|=0x30;
        videoint(0,0x07);
        memcpy((byte *)(mono_screen_buffer),save,4000);
        free(save);
        write_hardware_cursor(mono_crt_port,cursor);
      }
      init_current_screen();
      restore_bios_screen_state(&s);
    #else
      current_screen=other_screen;
      init_current_screen();
    #endif
    other_main_window=&(masters[current_screen]->w);
    current_screen=this_screen;
  }
  if (!lvideo_linestack) lvideo_linestack=create_linestack(16);
}

void copyline(int dest_y, int source_y)
{
  memcpy(wadr_of(0,dest_y),wadr_of(0,source_y),current_window->wrowsize*2);
}

static _clreol(int x, int y)
{
  word *c=wadr_of(x,y);
  word pattern;
  pattern = ( ( x ? *(c-1) : *c ) & 0xff00) | ' ';
  x=current_window->wrowsize-x;
  for (; x>0 ;x--) *(c++)=pattern;
}


char *cputs_limited(char *s, int count, int border)
{
  int l, c, sr=current_window->wrowsize;
  word *sb=current_window->regenbuf;
  if (((border << 1) +2 ) > current_window->wrowsize) border= (current_window->wrowsize -2) >> 1;
  current_window->wrowsize-=border<<1;
  current_window->regenbuf=&current_window->regenbuf[border];
  l=strlen(s);
  if (l>count) {
    c=s[count];
    s[count]=0;
  }
  cputs(s);
  current_window->wrowsize=sr;
  current_window->regenbuf=sb;
  if (l>count) {
    s[count]=c;
    return &s[count];
  }
  return &s[l];
}

void cputs_words(char *s, int border)
{
  int mx,l,lf;
  char *s_;
  while ((l=strlen(s)) != 0) {
    mx=wherex()-1 + (current_window->wrowsize-(border <<1));

    s_= strchr(s,'\n');
    if (s_) l=s_-s+1;
    if (l > mx) {
      l=mx-1;
      while ((s[l] != ' ') && (l) ) l--;
      if (!l) {
        /* giving up, too long */
        s=cputs_limited(s,mx,border);
        continue;
      }
      s=cputs_limited(s,l,border)+1;
      cputs("\n");
      continue;
    }
    s=cputs_limited(s,l,border);
  }
}

void scroll_down(int y)
{
  /* scrolls down the line at "y", inserts a blank line at "y" */
  int i;
  for (i=current_window->y2; i > y ; i--) copyline(i,i-1);
  _clreol(0,y);
}

void scroll_up(int y)
{
  /* scrolls up the line at "y", inserts a blank line at "y" */
  int i;
  if (current_window->full_screen) {
    memcpy(current_window->regenbuf,wadr_of(0,1),current_window->wrowsize*2*y);
  }
  else for (i=0; i<y; i++) copyline(i,i+1);
  _clreol(0,y);
}

void save_window_image(windowtyp *w)
{
  int size;
  if (w->framewindow) w=(windowtyp *)w->framewindow;
  size=w->wrowsize*w->wcolsize*2;
  if (!w->saveptr) w->saveptr=malloc(size);
  if (w->full_screen) {
    memcpy(w->saveptr,w->regenbuf,size);
  }
  else {
    int i;
    word *p;
    for (i=0,p=w->saveptr; i < w->wcolsize;  i++,p+=w->wrowsize) {
      memcpy(p,w->regenbuf+(i*w->rowsize),w->wrowsize*2);
    }
  }
}

void restore_window_image(windowtyp *w)
{
  int size;
  if (w->framewindow) w=(windowtyp *)w->framewindow;
  if (!w->saveptr) return;
  size=w->wrowsize*w->wcolsize*2;
  if (w->full_screen) {
    memcpy(w->regenbuf,w->saveptr,size);
  }
  else {
    int i;
    word *p;
    for (i=0,p=w->saveptr; i < w->wcolsize;  i++,p+=w->wrowsize) {
      memcpy(w->regenbuf+(i*w->rowsize),p,w->wrowsize*2);
    }
  }
}

int swap_window_image(windowtyp *w)
{
                   /* returns "false" if no image saved */
  int i;
  word *p;
  word aux[132];

  if (w->framewindow) w=(windowtyp *)w->framewindow;
  if (!w->saveptr) return (0);
  for (i=0,p=w->saveptr; i < w->wcolsize;  i++,p+=w->wrowsize) {
    memcpy(aux,w->regenbuf+(i*w->rowsize),w->wrowsize*2);
    memcpy(w->regenbuf+(i*w->rowsize),p,w->wrowsize*2);
    memcpy(p,aux,w->wrowsize*2);
  }
  return 1;
}

int reposition_window(windowtyp *w, int x1, int y1)
{
  screentyp *s=w->screen;
  int dx,dy;
  if (w->framewindow) {
    if (!reposition_window(w->framewindow,x1-1,y1-1)) return 0;
  }
  if ( (x1<0) || (y1<0) ) return 0;
  dx=x1 - w->x1;
  dy=y1 - w->y1;
  if ( ( (w->x2 + dx) > s->w.x2 ) || ( (w->y2 + dy) > s->w.y2 ) ) return 0;
  w->x1 += dx;
  w->y1 += dy;
  w->x2 += dx;
  w->y2 += dy;
  w->regenbuf=(s->w.regenbuf + (w->x1) + (w->y1*w->rowsize));
  return 1;
}

int move_window(windowtyp *w, int x1, int y1)
{
  int r;
  if (!swap_window_image(w)) return 0;
  r=reposition_window(w,x1,y1);
  swap_window_image(w);
  if (r && (current_window==w)) update_cursor(w);
  return r;
}


void make_window_frame(windowtyp *w,char *framepattern,int frameattrib)
{
  /* frame string must have format: "Ŀٳ"
                                     0123456  */
  int i;
  unsigned char *frame=(unsigned char *)framepattern;
  union {
    byte b[2];
    word w;
  } c;
  word *p;

  if (!frameattrib) frameattrib=(w->frameattrib>>8) & 255;
  if (w->framewindow) w=(windowtyp *)w->framewindow;
  c.b[1]=frameattrib;
  c.b[0]=0;
  *(w->regenbuf)=c.w|frame[0];
  *(w->regenbuf+w->wrowsize-1)=c.w|frame[2];
  *_wadr_of(w,0,w->wcolsize-1)=c.w|frame[3];
  *_wadr_of(w,w->wrowsize-1,w->wcolsize-1)=c.w|frame[5];
  c.b[0]=frame[1];
  for (i=w->wrowsize-2,p=w->regenbuf+1; i>0; i--) *(p++)=c.w;
  c.b[0]=frame[4];
  for (i=w->wrowsize-2, p=_wadr_of(w,1,w->wcolsize-1); i>0; i--) *(p++)=c.w;
  c.b[0]=frame[6];
  for (i=w->wcolsize-2, p=_wadr_of(w,0,1); i>0; i--,p+=w->rowsize) *(p)=c.w;
  for (i=w->wcolsize-2, p=_wadr_of(w,w->wrowsize-1,1); i>0; i--,p+=w->rowsize) *(p)=c.w;
}

void gotoxy(int x, int y)
{
  if (x<1) x=1;
  else if (x>current_window->wrowsize) x=current_window->wrowsize;
  if (y<1) y=1;
  else if (y>current_window->wcolsize) y=current_window->wcolsize;
  current_window->x=x-1;
  current_window->y=y-1;
  update_cursor(current_window);
}

int wherex()
{
  update_xy(current_window);
  return current_window->x+1;
}

int wherey()
{
  update_xy(current_window);
  return current_window->y+1;
}

void wherexy(int *x, int *y)
{
  update_xy(current_window);
  *x=current_window->x+1;
  *y=current_window->y+1;
}

void clearline(int dest_y)
{
  word *w=wadr_of(0,dest_y);
  word pattern=current_window->currattrib | ' ';
  int i=current_window->wrowsize;
  while (i--) *(w++)=pattern;
}

void clrscr(void)
{
  int i;
  if (current_window->full_screen) {
    word *w=current_window->regenbuf;
    word pattern=current_window->currattrib | ' ';
    int i=current_window->wrowsize*current_window->wcolsize;
    while (i--) *(w++)=pattern;
  }
  else for (i=0; i < current_window->wcolsize; i++) clearline(i);
  gotoxy(1,1);
}


void clreol()
{
  _clreol(current_window->x,current_window->y);
}

int cputs(char *str)
{
  word *p;
  int x,y,x2,y2;
  union {
    byte b[2];
    word w;
  } c;
  int ch;

  update_xy(current_window);
  x=current_window->x;
  y=current_window->y;
  x2=current_window->wrowsize-1;
  y2=current_window->wcolsize-1;
  p=wadr_of(x,y);
  c.w = current_window->currattrib;
  while ((ch=*(str++)) !=0) {
    switch (ch) {
      case 13: {
        x=0;
        p=wadr_of(x,y);
        break;
      }
      case 8: {
        if (x) {
          x--;
          p--;
        }
        break;
      }
      case 10: {
        x=0;
        if (y>=y2) {
          y=y2;
          scroll_up(y);
        }
        else y++;
        p=wadr_of(x,y);
        break;
      }
      default: {
        c.b[0]=ch;
        *(p++)=c.w;
        x++;
        if (x>x2) {
          x=0;
          if (y>=y2) {
            y=y2;
            scroll_up(y);
          }
          else y++;
          p=wadr_of(x,y);
        }
      }
    }
  }
  current_window->x=x;
  current_window->y=y;
  update_cursor(current_window);
  return ch;
}


int cprintf(char * format, ... )
{
  int count;
  va_list argptr;
  va_start(argptr, format);
  count=vsprintf(cprintf_buffer,format,argptr);
  cputs(cprintf_buffer);
  return count;
}

int cgetc_from_screen(int x,int y)
{
  return *(wadr_of(x-1,y-1)) & 0xff;
}

static int bounds(int min, int max, int value)
{
  if (value<min)  return min;
  else {
    if (value>max) {
      if (min<max) return max;
      else return min;
    }
    else return value;
  }
}

/* key-board package Ŀ */


#ifdef __bios_switch__

  static long int16call(int mode)
  {
    word scode;
    enum {ZFLAG=0x40};
    struct REGPACK r;
    r.r_ax=mode << 8;
    intr(0x16, &r);
    if (mode) {
      if (r.r_flags & ZFLAG) return 0;
      return 0xffff0000 | r.r_ax;
    }
    return r.r_ax;
  }

#else

  #define biosword(w) (*((word *)(bios_data+(w))))

  static long int16call(int mode)
  {
    enum {
      key_rptr=0x1a,
      key_wptr=0x1c,
      key_bufstart=0x80,
      key_bufend=0x82
    };
    word scode;
    if (!mode) {  /* wait for a keystroke */
      while (biosword(key_rptr)==biosword(key_wptr)) ;
      scode= biosword(biosword(key_rptr));
      if ((biosword(key_rptr)+2)==biosword(key_bufend)) biosword(key_rptr)=biosword(key_bufstart);
      else biosword(key_rptr)+=2;
    }
    else {      /* check for a keystroke */
      if (biosword(key_rptr)==biosword(key_wptr)) return 0;
      scode= biosword(biosword(key_rptr)) | 0xFFFF0000;
    }
    return scode;
  }
#endif

longint l_key_status()
{
  word scode;
  long lcode;
  lcode=int16call(1);
  if (!lcode) return 0;
  scode = lcode;
  if (scode & 0xff) return (scode & 0xff) | 0xFFFF0000;
  else return (scode>>8)  | 0xFFFF0100;
}

int l_key_input()
{
  word scode;
  scode=int16call(0);
  if (scode & 0xff) return (scode & 0xff);
  else return (scode>>8) | 0x100;
}

int check_for_keystroke(char *pattern)
{
  int c=l_key_status();
  char *s;
  while (c) {
    c=l_key_input();
    if (c < 256) {
      if (pattern) s=strchr(pattern,c);
      else s=(char *)1;
      if (s) return c;
    }
  }
  return c;
}


/* key-board package  */

/* string and other input Ŀ*/

#define line_max    256
typedef struct{
  int stack_max, stack_max_;
  int x,x_,xmax,y;
  int wri,rdi;
  int single;
  char line[line_max+1];
  char s[2] [line_max+1];
} linestacktyp;


void *create_linestack(int numlines)
{
  linestacktyp *stck;
  int size =  sizeof(linestacktyp)+((numlines-2)*(line_max+1));
  stck=malloc(size);
  memset(stck,0,size);
  stck->stack_max = numlines;
  return stck;
}

static void inc_linestack(linestacktyp *stck,int *i)
{
  (*i)++;
  if (*i >= stck->stack_max_) *i=0;
}
static void dec_linestack(linestacktyp *stck,int *i)
{
  (*i)--;
  if (*i < 0) *i = stck->stack_max_-1;
}

static void flushline(linestacktyp *stck)
{
  int j,max,x2;
  gotoxy(stck->x,stck->y);
  x2=current_window->wrowsize;
  cputs(stck->line);
  max=strlen(stck->line);
  for (j=max; j<stck->xmax; j++) cputs(" ");
  stck->xmax=max;
  if ( (stck->x_ - stck->x) > max)  stck->x_ = stck->x+ max;
  if (stck->x_ <= x2) gotoxy(stck->x_,stck->y);
  else {
    gotoxy(stck->x_ % x2,
     bounds(stck->y, stck->y+current_window->wcolsize-1 , stck->y+(stck->x_ / x2)));
  }
}

static void  movecursor(linestacktyp *stck,int xx, int yy, int topline)
{
  if (yy !=0) {
    if (stck->single) {
      stck->x_ += current_window->wrowsize * yy;
      stck->x_=bounds(stck->x,stck->x + strlen(stck->line) + 1,stck->x_+xx);
    }
    else {
      if (stck->rdi == topline) strcpy(stck->s[topline],stck->line);
      switch (yy) {
        case -1: dec_linestack(stck,&stck->rdi); break;
        case +1: inc_linestack(stck,&stck->rdi); break;
      }
      strcpy(stck->line,stck->s[stck->rdi]);
      stck->x_ = bounds(stck->x, stck->x + strlen(stck->line) + 1, stck->x_+255);
      flushline(stck);
    }
  }
  else stck->x_=bounds(stck->x,stck->x + strlen(stck->line) + 1,stck->x_+xx);
}

static void insert(char *s,int pos,char c)
{
   char cc,ch;
   if (*s){
   int len=strlen(s);
   if (pos>len) pos=len;
   if (pos>=line_max) return;
   s+=pos;
   cc=*s;
   *s++=c;
   while (cc) {
     ch=cc;
     cc=*s;
     *s=ch;
     s++;
   }
   *s=cc;
   }
   else {
     *(s++)=c;
     *s=0;
   }
}

static void delete(char *s,int pos)
{
  int len=strlen(s);
  if ((pos<0) || (pos>=len)) return;
  strcpy(&s[pos],&s[pos+1]);
}

void cpush_onto_linestack(void *line_stack, char *s)
{
  linestacktyp *stck=line_stack;
  int len=strlen(s);
  if (len >= line_max) len = line_max-1;
  strncpy(stck->s[stck->wri],s,len);
  stck->s[stck->wri][len]=0;
  strncpy(stck->line,s,len);
  stck->line[len]=0;
  stck->xmax=len+1;
  if ( stck->stack_max_ < stck->stack_max ) {
    stck->stack_max_++;
    stck->wri++;
  }
  else inc_linestack(stck,&stck->wri);
}


void cstring_input(void *line_stack, char *ss, char *source)
{
  int c, wri_;
  linestacktyp *stck=line_stack;

  wherexy(&stck->x, &stck->y);
  stck->x_=stck->x;
  if (source) cpush_onto_linestack(stck,source);
  else cpush_onto_linestack(stck,"");
  stck->rdi=stck->wri;
  dec_linestack(stck,&stck->rdi);
  wri_=stck->rdi;
  flushline(stck);
  do {
    c=l_key_input();
    switch (c) {
      case CursorUp:    movecursor(stck,0,-1,wri_); break;
      case CursorDown:  movecursor(stck,0, 1,wri_); break;
      case CursorLeft:  movecursor(stck,-1,0,wri_); break;
      case CursorRight: movecursor(stck, 1,0,wri_); break;
      case DEL_key:{
        delete(stck->line,stck->x_-stck->x);
        break;
      }
      case KeyHome: {
        stck->x_=stck->x;
        break;
      }
      case KeyEnd: {
        stck->x_=stck->xmax+1;
        break;
      }
      case 8:{
        delete(stck->line,stck->x_-stck->x-1);
        movecursor(stck,-1,0,wri_);
        break;
      };
      case ESC_key:{
        ss[0]=0;
        return;
      };
      default: {
        if ( (c>=' ') && (c!=0x7f) ) {
          if ((stck->xmax < (line_max-1)) && ((stck->x_-stck->x) < (line_max-1))) {
            insert(stck->line,stck->x_-stck->x,c);
            movecursor(stck, 1,0,wri_);
          }
        }
        break;
      }
    }
    flushline(stck);
  } while (c!=13);
  strcpy(stck->s[wri_],stck->line);
  strcpy(ss,stck->line);
}

char *cgetstring(char *ss)
{
  cstring_input(lvideo_linestack,ss,NULL);
  return ss;
}

char *cupdatestring(char *ss)
{
  cstring_input(lvideo_linestack,ss,ss);
  return ss;
}

int cchange_attrib(int x1, int y1, int x2, int y2, int attr)
{
  int xsize,y,attrold;
  char *b,*bmax;

  if (x1>x2) {y=x1; x1=x2; x2=y;}
  if (y1>y2) {y=y1; y1=y2; y2=y;}
  x1--; y1--;
  xsize=2*(x2-x1);
  attrold=*(((char *)(wadr_of(x1,y1)))+1);
  for (y=y1; y<y2; y++) {
    b=((char *)(wadr_of(x1,y)))+1;
    for (bmax=b+xsize; b<bmax; b+=2) *b=attr;
  }
  return attrold;
}

int cmenu_bar_input(int x1, int y1, int x2, int y2, int xread, int yfirst)
{
   int oldattr;
   int y_,c,loop;
   int y=yfirst;
   windowtyp *w = current_window;
   loop=1;

   do {
     oldattr=cchange_attrib(x1,y,x2,y,(w->inversattrib >> 8) & 0xff);
     gotoxy(x1,y);
     y_=y;
     c=l_key_input();
     switch (c) {
       case PgUp:       y=y1; break;
       case CursorUp:   y=bounds(y1,y2,y-1); break;
       case PgDn:       y=y2; break;
       case CursorDown: y=bounds(y1,y2,y+1); break;
       case Enter:      c=*(wadr_of(xread-1,y-1)) & 0xff; loop=0; break;
       default:         loop=0; break;
     }
     cchange_attrib(x1,y_,x2,y_,oldattr);
   } while (loop);
   return c;
}


/*
 * this creates a pointer list of individual string lines
 * contained in "list", '\n'is the separator.
 */
char **make_stringlist(char *list, int start, int step)
{
  enum { maxi=128 };
  int i,j,count,size;
  char **sl_;
  char *sl[maxi];
  char *s;

  s=list-1;
  count=0;
  do {
    sl[count++]=++s;
  } while ((count < maxi) &&  ((s=strchr(s,'\n')) !=0)) ;
  if (!sl[count-1][0]) count--;

  size=((((count-start)+step-1)/step)+1)*sizeof(void *);
  sl_=malloc(size);
  for (i=start, j=0; i<count; i+=step,j++) sl_[j]=sl[i];
  sl_[j]=0;
  return sl_;
}

/* NOTE: This is used to replace all '\n' in a list (preveously genarated
   by make_stringlist) with '\0' and later restore it again.
   USAGE:
     int save;
     save= replace_stringlist_terminator(list, count, 0);
     ... do some thing with the list ...
     replace_stringlist_terminator(list, count, save);
 */
int replace_stringlist_terminator(char **list, int count, int newtermin)
{
  int i,l;
  for (i=1; i<count; i++) {
    list[i][-1]=newtermin;
  }
  l=strlen(list[count-1]);
  if (newtermin) {
    newtermin >>=8;
    if (newtermin) list[count-1][l]='\n';
  }
  else {
    if (l && (list[count-1][l-1]=='\n')) {
      list[count-1][l-1]=0;
      return '\n'<<8;
    }
    return '\n';
  }
}

static  _cputslist(char **list, int x1, int y1, int x2, int y2)
{
  int i,count;
  char *s,c;
  gotoxy(x1,y1);
  count=x2-x1+1;
  for (i=0; i<= (y2-y1); i++) {
    gotoxy(x1,y1+i);
    if (!list[i]) break;
#if 0
    cputs(list[i]);
#else
    if ((s=strchr(list[i],'\n')) !=0) {
      c=*s;
      *s=0;
    }
    if (    ((y1+i)==current_window->wcolsize)
         && ((x1+count) >= current_window->wrowsize) )
         cputs_limited(list[i], count-1, 0);
    else cputs_limited(list[i], count, 0);
    if (s) *s=c;
#endif
    while (wherex()<x2) cputs(" ");
  }
}

int _get_list_length(char **list)
{
   int maxli=0;
   char **s=list;
   while(s[0]) {maxli++; s++;}
   return maxli;
}

static int _get_max_length_from_list(char **list)
{
   int maxl=0,i;
   char **s=list;
   while(s[0]) {
     i=0;
     while (s[0][i] && (s[0][i]!='\n') ) i++;
     if (i>maxl) maxl=i;
     s++;
   }
   return maxl;
}

int cmenu_list_input(char **list,
    int x1, int y1, int x2, int y2, int yfirst,
    cmenu_callbacktyp *callback, void *callbackparams)
{
   int oldattr;
   int y_,c,loop=1;
   int y=yfirst+1;
   int li=0, li_, lastli=-1, maxli, i;
   char **s;
   windowtyp *w = current_window;
   x1++,y1++,x2++,y2++;

   loop=1;

   maxli=_get_list_length(list)-1;
   if (y1+maxli < y2) y2=y1+maxli;
#if 1
   li_=bounds(0,maxli,cmenu_choice_preset);
   cmenu_choice_preset=-1;
   y=(yfirst+1)+li_;
   if (y>y2) {
     li=li_;
     y=y1;
     if ((li+(y2-y1)) > maxli) {
       li=maxli-(y2-y1);
       y=y1+(li_-li);
     }
   }
#else
   if (y>y2) y=y2;
#endif
   _cputslist(list+li, x1, y1, x2, y2);

   do {
     if (callback) {
       if ((li+y) !=lastli) {
         if ( (*((cmenu_callbacktyp *)callback))(li+(y-y1),callbackparams) ) return li+(y-y1);
         lastli=li+y;
       }
     }
     oldattr=cchange_attrib(x1,y,x2,y,(w->inversattrib >> 8) & 0xff);
     gotoxy(x1,y);
     y_=y;
     c=l_key_input();
     switch (c) {
       case PgUp:
         y=y1; li=1;
         /* fall through */
       case CursorLeft:
       case CursorUp: {
         if (y<=y1) {
           if (li > 0) {
             cchange_attrib(x1,y_,x2,y_,oldattr);
             oldattr=-1;
             li--;
             _cputslist(list+li, x1, y1, x2, y2);
           }
         }
         else y--;
         break;
       }
       case PgDn:
         y=y2; li=maxli-(y2-y1)-1;
         /* fall through */
       case CursorRight:
       case CursorDown: {
         if (y>=y2) {
           if (li+(y-y1) < maxli) {
             cchange_attrib(x1,y_,x2,y_,oldattr);
             oldattr=-1;
             li++;
             _cputslist(list+li, x1, y1, x2, y2);
           }
         }
         else y++;
         break;
       }
       case Enter:      c=li+(y-y1); loop=0; break;
       case ESC_key:    c=-1; loop=0; break;
       default:         break;
     }
     if (oldattr>=0) cchange_attrib(x1,y_,x2,y_,oldattr);
   } while (loop);
   return c;
}

static windowtyp *create_button_box(char *text, int x, int y, int size, int frameattrib)
{
  windowtyp *w, *w_;
  char *s, c;
  char *framepattern="Ŀٳ";
  int l;
  w_=current_window;
  if (!frameattrib) frameattrib=(w_->frameattrib>>8) & 255;
  w=createframewindow(this_screen, 0, x, y, x+size, y+2);
  make_window_frame(w,framepattern,frameattrib);
  goto_window(w);
  clrscr();

  size-=2;
  if ((s=strchr(text,'\n')) !=0) { c=*s; *s=0; }
  l=strlen(text);
  if (l>size) l=size;
  gotoxy(((size+1)/2)-((l+1)/2)+1,1);
  cputs_limited(text, size, 0);
  if (s) *s=c;
  goto_window(w_);
  return w;
}

int cmenu_button_input(char **list, int x1, int y1, int x2, int y2, char *framepattern, int frameattrib1, int frameattrib2)
{
  windowtyp *w, *w_, *bw[10];
  int withcursor=0, maxli;
  int bsize,bx,i,li,li_,oldattr,c,loop=1;
  int border=4;
  int buttonsize;

  w_=current_window;
  if (!frameattrib1) frameattrib1=(w_->frameattrib>>8) & 255;
  goto_window(this_main_window);
  w=createframewindow(this_screen, withcursor, x1, y1, x2, y2);
  save_window_image(w);
  make_window_frame(w,framepattern,frameattrib1);
  goto_window(w);
  clrscr();
  goto_window(this_main_window);
  maxli=_get_list_length(list);
  buttonsize= _get_max_length_from_list(list)+4;

  bsize=(((x2-x1)-2*border)+(maxli/2))/maxli;
  for (i=0,bx=((bsize+1)/2)-((buttonsize+1)/2); i<maxli; i++, bx+=bsize) {
    bw[i]=create_button_box(list[i],x1+bx+border,y1+(y2-y1)/2 -1, buttonsize, frameattrib2);
  }


  li=bounds(0,maxli-1,cmenu_choice_preset);
  cmenu_choice_preset=-1;
  do {
    goto_window(bw[li]);
    oldattr=cchange_attrib(1,1,bw[li]->wrowsize,1,(w->inversattrib >> 8) & 0xff);
    li_=li;
    c=l_key_input();
    switch (c) {
      case PgUp:
      case CursorLeft:
      case CursorUp: {
        if (li <=0) li=maxli-1;
        else li--;
        break;
      }
      case PgDn:
      case CursorRight:
      case CursorDown: {
        if (li >= (maxli-1)) li=0;
        else li++;
        break;
      }
      case Enter:      c=li+1; loop=0; break;
      case ESC_key:    c=0; loop=0; break;
      default:         break;
    }
    if (oldattr>=0) cchange_attrib(1,1,bw[li_]->wrowsize,1,oldattr);
  } while (loop);


  restore_window_image(w);
  goto_window(w_);
  for (i=0; i<maxli; i++) {
    free_window(bw[i]);
  }
  free_window(w);
  return c;
}

void cmenu_line_input(char *dest, char *thisline, char **list,
   int x1, int y1, int x2, int y2, char *framepattern, int frameattrib)
{
  windowtyp *w, *w_, *bw[10];
  int withcursor=1, maxli;
  int bsize,bx,i,li,li_,oldattr,c;
  linestacktyp *stck;

  w_=current_window;
  if (!frameattrib) frameattrib=(w_->frameattrib>>8) & 255;
  goto_window(this_main_window);
  w=createframewindow(this_screen, withcursor, x1, y1, x2, y2);
  save_window_image(w);
  make_window_frame(w,framepattern,frameattrib);
  goto_window(w);
  w->currattrib=w->highlightattrib;
  clrscr();

  if (list && list[0]) {
    maxli=_get_list_length(list);
    stck=create_linestack(maxli+1);
    for (i=0; i<maxli; i++) cpush_onto_linestack(stck,list[i]);
  }
  else {
    maxli=1;
    stck=create_linestack(maxli+1);
    stck->single=1;
  }

  cstring_input(stck,dest,thisline);

  restore_window_image(w);
  goto_window(w_);
  free_window(w);
  free(stck);
}



/**/



#if 0    /* only for test purposes */

char *xxlist[] ={
  "1xxx",
  "2x_\nxxxxxxxx",
  "3xxxxxxxxxxx",
  "4xxxxxxxxxxxxxx",
  "5xxxxxxxx",
  "6xxxxxxxxxx",
  "7xxxx",
  "8xxxxxxx",
  0
};
char yylist[]="11111\n222\n333333333\n444444\n5555\n66666\n77777\n888";

static int testclist_callback(int line,void *dymmy)
{
  static count=0;
  gotoxy(1,25);
  count++;
  if (count>50) return 1;
  cprintf("%d. callback, beeing on item %d of list_choice",count,line);
  return 0;
}
static void testclist()
{
  char **list;
  clrscr();
  list=make_stringlist(yylist,1,3);
  printf("\n\nreturns line %d\n",
    cmenu_list_input(list,80-20,10,80-3,25,15,testclist_callback,0));
}

static char *butlist[]={
  "aaaa","bbbccbbb", "Cancel",0, "dd",0
};

void testbuttons()
{
   gotoxy(1,25);
   cprintf("return %d",cmenu_button_input(butlist, 10,10,  80-2*10+1,14, "Ŀٳ",0,0));
}

main()
{
  screentyp s;
  windowtyp *w1;
  char ss[128];
  char c;
  int x,y;
  void *stck;

  save_bios_screen_state(&s);
  printf("has_HGC=%d  has_EGA=%d\n",check_for_herkules(),check_for_egavga());
  initvideo();

#if 0
  testbuttons();
  exit(0);
#endif

#if 0
  testclist();
  exit(0);
#endif

#if 1
  stck=create_linestack(4);
  cpush_onto_linestack(stck,"111111");
  cpush_onto_linestack(stck,"2222");
  cstring_input(stck,ss,"3333333");
  cprintf("\nafter cstring_input>%s<\n",ss);
  exit(0);
#endif


#if 0
  do {
    cgetstring(ss);
    cputs("\n");
  } while (*ss);
  exit(0);
#endif

  cputs("abcdefghijk1111\n222222222\n3333333333\n");
  wherexy(&x,&y);
  c=cmenu_bar_input(x,y-3,x+20,y-1,x+2,y-1);
  gotoxy(x,y);
  cprintf("inputed character was >%c<\n",c);
  exit(0);

  goto_window(other_main_window);
  cputs("aaaaaaaaaaaaaaa\nbbbbbbbbb\ncccccccccc\n");
  goto_window(this_main_window);
  cprintf("hello world %d %d\n",1,2);
  w1=createframewindow(this_screen,1,40,12,60,20);
  w1->currattrib=w1->inversattrib;
  goto_window(w1);

  save_window_image(w1);
  clrscr();
  cputs("xxxxxxxxxxxxxxxxxxxxxx\nyyyyyyyyyyyyy\nzzzzzzzzzz\n");
  make_window_frame(w1,"ͻͼ",0x1f);
  l_key_input();
  move_window(w1,w1->x1-8,w1->y1-4);
  l_key_input();
  move_window(w1,w1->x1-8,w1->y1-4);
  l_key_input();
  cputs("1234123412341234123412\nabcdefghijklmn\nABCDEFGHIJ\n");
  l_key_input();
  goto_window(this_main_window);
  exit(0);
}

#endif
