/* This is file LVIDEO.H
   Version 2.0
============================================================================

   (C) 1992..1995 Hans Lermen, D5480 Remagen, GERMANY
                  (lermen@elserv.ffm.fgan.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


============================================================================

   Direct dual-monitor video package,
   without (or a minimum) of PC-BIOS-calls.
   
   If __bios_switch__ is not defined, then NONE bios calls is done!
   All is done by directly accessing the only well known PC hardware.
   This is faster and (more important) avoids BIOS incompatibilities
   and last not least can be run by other CPU then the ix86 family
   (i.e. i860, 68K, and others on an ISA/EISA architecture )

   This Package has been successful compiled by the following C-Compiler:

   TURBO-C  (Borland)
   DJGCC    (Free Software Foundation, D.J.Delorie's DOS-portation)
   HIGH-C   (MetaWare, i860 Compiler)

*/

#ifndef __LVIDEO_H__
#define __LVIDEO_H__

#ifdef __cplusplus
extern "C" {
#endif

#ifndef __dword__
#define __dword__
typedef unsigned long dword;
typedef long longint;
typedef unsigned short word;
typedef unsigned char byte;
#endif

#define bios_xpos (*((byte *)(bios_data+0x50)))
#define bios_ypos (*((byte *)(bios_data+0x51)))
#define wadr_of(x,y) (current_window->regenbuf + (x) + ((y)*current_window->rowsize))
#define _wadr_of(w,x,y) ((w)->regenbuf + (x) + ((y)*(w)->rowsize))


typedef struct {
  word *regenbuf;  /* screen address of upper left corner of window */
  void *screen;    /* " screentyp * " */
  void *framewindow; /* pointer to window which defines a frame around */
  word *saveptr;   /* pointer to screen-image save area (on the heap!) */
  int what_screen; /* can be "this_screen" or "other_screen" */
  int with_cursor; /* !=0, if with hardware-cursor update  */
  int full_screen; /* !=0, if windows equals to full-screen */
  int wrowsize;    /* size in words of a row in the window */
  int wcolsize;    /* number of columns in the window */
  int rowsize ;    /* size in words of a row on the screen */
  int x,y;         /* koordinates (0...) of the current cursor position
                      relative to left upper corner of the window */
  int x1,y1,x2,y2; /* koordinates (0..._ of the left upper (x1,y1) and
                      the rigth lower (x2,y2) corner of the window
                      relative to left upper corner of the screen */
                   /* Attribute templates:
                      bits 8..15 = attribute byte, all other bits ZERO */
  int currattrib;      /* current attribute */
  int normattrib;      /* Attribute for normal text */
  int inversattrib;    /* Attribute for invers text */
  int highlightattrib; /* Attribute for highlighted text */
  int frameattrib;     /* Attribute for frame of frame window */
} windowtyp;

typedef struct {
  windowtyp w;
  word *regenbuf; /* address of upper left corner of the regen-buffer (screen)*/
  int crt_index,crt_data;
  int equipflag;         /* 40:10 */
       byte xxx1;        /* align structure to ODD to avoid padding below */
       /*  part 1 Ŀ */
  byte current_mode;     /* 40:49 */
  word numcolumns,       /* 40:4a */
       screenpagesize,   /* 40:4c */
       currentpagestart; /* 40:4e */
  byte cursors[16];      /* 40:50 */
  word cursortype;       /* 40:60 */
  byte currentpagenum,   /* 40:62 */
       crtbase_low,crtbase_high,/* 40:63 */
       modeselectreg,    /* 40:65 */
       palettereg;       /* 40:66 */
       /*  part 1  */
       /*  part 2 Ŀ */
  byte egavga_params[7];  /* 40:84 */
       /*  part 2  */
       /*  part 3 Ŀ */
  byte egapointer[4];     /* 40:a8 */
       /*  part 3  */
} screentyp;


#ifdef LVIDEO_itself
  #define LVIDEO_PUBLIC
#else
  #define LVIDEO_PUBLIC extern
#endif

enum { no_screen=0, this_screen=1,  other_screen=2 };

LVIDEO_PUBLIC int color_screen, mono_screen;
LVIDEO_PUBLIC int screen_in_graphics_mode;
LVIDEO_PUBLIC int dual_monitor;
LVIDEO_PUBLIC int cmenu_choice_preset;
#ifdef LVIDEO_itself
  int current_screen=this_screen;
  windowtyp *this_main_window=0;
  windowtyp *other_main_window=0;
  windowtyp *current_window=0;
  void *lvideo_linestack=0;
#else
  LVIDEO_PUBLIC int current_screen;
  LVIDEO_PUBLIC windowtyp *this_main_window;
  LVIDEO_PUBLIC windowtyp *other_main_window;
  LVIDEO_PUBLIC windowtyp *current_window;
  LVIDEO_PUBLIC void *lvideo_linestack;
#endif


int check_for_egavga();
int check_for_herkules();
int check_for_graphics_mode();
void update_cursor(windowtyp *w);
void update_xy(windowtyp *w);
void goto_screen(int what_screen);
void goto_window(windowtyp *w);
windowtyp *createwindow(int screen, int withcursor, int x1, int y1, int x2, int y2);
void free_window(windowtyp *w);
windowtyp *createframewindow(int screen, int withcursor, int x1, int y1, int x2, int y2);
void initvideo();
void set_global_attributes(int screen, int normattrib,
           int inversattrib, int highlightattrib, int frameattrib);
void copyline(int dest_y, int source_y);
void scroll_down(int y);
void scroll_up(int y);
void save_window_image(windowtyp *w);
void restore_window_image(windowtyp *w);
int swap_window_image(windowtyp *w);
int reposition_window(windowtyp *w, int x1, int y1);
int move_window(windowtyp *w, int x1, int y1);
void make_window_frame(windowtyp *w,char *framepattern,int frameattrib);
void gotoxy(int x, int y);
int wherex();
int wherey();
void wherexy(int *x, int *y);
void clearline(int dest_y);
void clrscr(void);
void clreol();
int cputs(char *str);
char *cputs_limited(char *s, int count, int border);
void cputs_words(char *s, int border);
int cprintf(char * format, ... );
int cgetc_from_screen(int x,int y);

void *create_linestack(int numlines);
char *cgetstring(char *ss);
char *cupdatestring(char *ss);
void cpush_onto_linestack(void *line_stack, char *s);
void cstring_input(void *line_stack, char *ss, char *source);

enum {
  BACKSPACE_key    = 8,
  TAB_key          = 9,
  Enter            = 13,
  ShiftTab         = 15,
  ESC_key          = 27,
  F1key            = (256+59),
  KeyHome          = (256+71),
  CursorUp         = (256+72),
  PgUp             = (256+73),
  CursorLeft       = (256+75),
  CursorRight      = (256+77),
  KeyEnd           = (256+79),
  CursorDown       = (256+80),
  PgDn             = (256+81),
  INS_key          = (256+82),
  DEL_key          = (256+83),
  CtrlCursorLeft   = (256+115),
  CtrlCursorRight  = (256+116),
  CtrlEnd          = (256+117),
  CtrlPgDn         = (256+118),
  CtrlHome         = (256+119),
  CtrlPgUp         = (256+132)
};
longint l_key_status();
int l_key_input();

/* return 0, if no keystroke in from pattern was entered.
   else returns the key.
   Does NOT block. If pattern = NULL, the any stroke returns != 0;
 */
int check_for_keystroke(char *pattern);


int cchange_attrib(int x1, int y1, int x2, int y2, int attr);
int cmenu_bar_input(int x1, int y1, int x2, int y2, int xread, int yfirst);

char **make_stringlist(char *list, int start, int step);
int _get_list_length(char **list);
int replace_stringlist_terminator(char **list, int count, int newtermin);
typedef int cmenu_callbacktyp(int line, void *params);
int cmenu_list_input(char **list,
    int x1, int y1, int x2, int y2, int yfirst,
    cmenu_callbacktyp *callback, void *callbackparams);
int cmenu_button_input(char **list, int x1, int y1, int x2, int y2, char *framepattern, int frameattrib1, int frameattrib2);
void cmenu_line_input(char *dest, char *thisline, char **list,
   int x1, int y1, int x2, int y2, char *framepattern, int frameattrib);

#ifdef __cplusplus
}
#endif

#endif

