%{
/*  >>> this is file parser.y
        use the GNU bison to generate parser.c
        parser.c should compile with GCC and with TURBO-C in large model
============================================================================
   LSETUP v1.0 (C) 1995 Hans Lermen (lermen@elserv.ffm.fgan.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


============================================================================
*/

#if 0
#define TESTPARSER
#endif

#ifndef __linux__
  #include "_malloc_.h" /* malloc, realloc intercept, must be placed TOP */
  #ifndef YYDEBUG
    #define YYDEBUG 0
  #endif
  #ifndef YYPURE
    #define YYPURE
  #endif
  #if 0
    #define STDERR stdout
  #endif
#else
  #define strdup our_strdup
#endif

#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
 #include <setjmp.h>
 #include <sys/stat.h>                    /* structure stat       */
#ifdef __linux__
 #include <unistd.h>                      /* prototype for stat() */
#endif
#ifndef __linux__
  #ifndef __TURBOC__
    #include <no_varg.h> /* undefine va_-stuff due to bad DJGCC-headers */
  #endif
#endif
#include <stdarg.h>

#ifdef STDERR
  /* need this on DOS, because we can't do 2>&1 */
  #define stderr STDERR
#endif
 
static int errors = 0;
static int warnings = 0;

	/* external procedures */
extern char* strdup(const char *); /* Not defined in string.h :-( */
extern int yylex(); /* exact argument types depend on the way you call bison */

	/* our definitions and structures */
#define __PARSER_itself
#include "parserd.h"

	/* local procedures */
static FILE* open_file(char* filename);
static void close_file(FILE* file);

static char *string_append(char *s1, char *s2);
static char *str_append(char *s1, char *s2);
static char *str_append2(char *s1, char *s2, char *s3);
static char *str_append3(char *s1, char *s2, char *s3, char *s4);
static long *list_append(long *list, long value);
static void *create_show_actiontyp(char *ident, char *menuname, struct objecttyp *image, void **action);
static void *create_waitkey_actiontyp(char *ident, char *possible_strokes);
static void *create_generic_actiontyp(char *ident, void *obj, int kind, int flags);
static void *create_call_actiontyp(char *ident,char *function,void *arg,
                                   int kind, int flags);
static void *create_set_actiontyp(char *ident, char *dest, char *src, int flags);
static void *create_menu_body(char *image, void *menuaction);
static void *create_action_body(void *head, void *tail);
static void create_menu_obj(char *mlabel, struct genobjecttyp **geom, void *body);


void yyerror(char *, ...);
void yywarn(char *, ...);

	/* variables in lexer.l */
extern int line_count;
extern FILE* yyin;
extern void yyrestart(FILE *input_file);
extern char *yy_vbuffer;
extern char * include_fnames[];
extern int include_stack_ptr;


#ifdef __linux__
char *strdup(const char *s)
{
  int l,l_;
  char *t;
  l=strlen(s)+1;
  if (l<9) l_=9;
  else l_=l;
  t=malloc(l_);
  memcpy(t,s,l);
  return t;  
}

#endif

%}

%start lines

%union {
  long int i_value;
  double r_value;
  char *s_value;
  void *p_value;
  };

%token <r_value> UREAL
%token <i_value> UINTEGER
%token <s_value> STRING IDENT IDENT2

%token START COLORS TSTRING TINTEGER TVOID 
%token MENU IMAGE ACTION FUNCTION SHOW WAITKEY PRINTXY SET
%token LABEL CALL GOTO IFERROR IFNOTERROR IFTRUE IFFALSE

%type <i_value> expression integer_expression
/*
%type <i_value>  INTEGER
%type <r_value> REAL
*/
%type <s_value> menustring optional_ident 
%type <s_value> expressionstring expstring
%type <p_value> menuimage arglist arglistlist arglistitem menubody
%type <p_value> menuaction actionbody action

%left '&'  '^' '|'
%left SHL_OP SHR_OP
%left '+' '-'
%left '*' '/' '%'
%left UCOMPL UMINUS

%%

lines:		line
		| lines line
		;

line:		second_pass 
		| IDENT ':' MENU arglist '{' menubody '}' {
		  create_menu_obj($1,$4,$6);
		}
		| IDENT ':' FUNCTION '{' actionbody '}' {
		  create_objekt($1,$5,  OBJ_FUNCTION);
		}
		| IDENT ':' TSTRING '{' menustring '}' {
		  if (!create_objekt($1,$5,OBJ_VARSTRING)) {
		    yyerror("'%s' multiple defined",$1);
		    free($1); free($5);
		  }
		}
		| IDENT ':' TSTRING {
		  char *s=strdup("0"); /* this because some systems return
		                          NULL on strdup(""), don't know why */
		  s[0]=0;
		  if (!create_objekt($1,s,OBJ_VARSTRING)) {
		    yyerror("'%s' multiple defined",$1);
		    free($1); free(s);
		  }
		}
		| IDENT ':' TVOID {
		  char *s=strdup("0");
		  s[0]=0;
		  if (!create_objekt($1,s,OBJ_VARSTRING)) {
		    yyerror("'%s' multiple defined",$1);
		    free($1); free(s);
		  }
		}

/* implemented only in parser
   NOT yet in setup, so better disable it.

		| IDENT ':' TINTEGER '{' expression '}' {
		  if (!create_objekt($1,(void *)$5,OBJ_VARINTEGER)) {
		    yyerror("'%s' multiple defined",$1);
		    free($1);
		  }
		}
		| IDENT ':' TINTEGER {
		  if (!create_objekt($1,0,OBJ_VARINTEGER)) {
		    yyerror("'%s' multiple defined",$1);
		    free($1);
		  }
		}
*/
		| START IDENT { create_objekt(strdup("start"),$2,OBJ_VARSTART); }
		| COLORS arglist { create_objekt(strdup("default_colors"),$2,OBJ_COLORS); }
		| error
		;

second_pass:	SET IDENT2 '=' expression {
		  char s[40];
		  sprintf(s,"%ld",$4);
		  set_string_by_name($2, s);
		  free($2);
		}
		;


menubody:
		menuimage {
		  $$=create_menu_body($1,0);
		}
		| menuimage menuaction {
		  $$=create_menu_body($1,$2);
                }
		;


menuimage:			
		IMAGE '{' menustring '}' { $$=create_objekt(0,$3,OBJ_VARSTRING); }
		| IMAGE IDENT {
		  $$=(void *)create_objekt(0,(void *)$2,OBJ_ARG_ID);
		}
		| IMAGE '{' IDENT '}' {
		  $$=(void *)create_objekt(0,(void *)$3,OBJ_ARG_ID);
		}
		;

menuaction:	ACTION '{' actionbody '}' { $$=$3; }
		;



actionbody:
		action {
		  $$=create_action_body($1,0);
		}
		|actionbody action {
		  $$=create_action_body($1,$2);
		}
		;


action:
		IDENT ':'  LABEL {
		  $$=create_generic_actiontyp($1,0,OBJ_NONE,0);
		}
		| optional_ident SHOW '(' IDENT ')' {
		  $$=create_show_actiontyp($1,$4,0,0);
		}
		| optional_ident SHOW '(' IDENT ',' arglistitem ')' {
		  $$=create_show_actiontyp($1,$4,$6,0);
		}
		| optional_ident SHOW '(' IDENT  ',' ',' arglistitem ')' {
		  $$=create_show_actiontyp($1,$4,0,$7);
		}
		| optional_ident SHOW '(' IDENT ',' arglistitem ',' arglistitem ')' {
		  $$=create_show_actiontyp($1,$4,$6,$8);
		}
		| optional_ident WAITKEY '(' STRING ')' {
		  $$=create_waitkey_actiontyp($1,$4);
		}
		| optional_ident WAITKEY '(' ')' {
		  $$=create_waitkey_actiontyp($1,0);
		}
		| optional_ident PRINTXY '(' arglistlist ')' {
                  $$=create_generic_actiontyp($1,$4,OBJ_PRINTXYACT,0);
		}
		| optional_ident CALL IDENT '(' arglistlist ')' {
                  $$=create_call_actiontyp($1,$3,$5,OBJ_CALLACT,0);
		}
		| optional_ident CALL IDENT '(' ')' {
                  $$=create_call_actiontyp($1,$3,0,OBJ_CALLACT,0);
		}
		| optional_ident IFTRUE GOTO IDENT {
                  $$=create_generic_actiontyp($1,$4,OBJ_GOTOACT,GOTO_IF_TRUE);
		}
		| optional_ident IFFALSE GOTO IDENT {
                  $$=create_generic_actiontyp($1,$4,OBJ_GOTOACT,GOTO_IF_FALSE);
		}
		| optional_ident IFERROR GOTO IDENT {
                  $$=create_generic_actiontyp($1,$4,OBJ_GOTOACT,GOTO_IF_TRUE | GOTO_IF_ERROR);
		}
		| optional_ident IFNOTERROR GOTO IDENT {
                  $$=create_generic_actiontyp($1,$4,OBJ_GOTOACT,GOTO_IF_FALSE | GOTO_IF_ERROR);
		}
		| optional_ident IFFALSE IDENT arglist {
                  $$=create_call_actiontyp($1,$3,$4,OBJ_GOTOACT,GOTO_CALL_FALSE);
		}
		| optional_ident IFTRUE IDENT arglist {
		  $$=create_call_actiontyp($1,$3,$4,OBJ_GOTOACT,GOTO_CALL_TRUE);
		}
		| optional_ident GOTO IDENT {
                  $$=create_generic_actiontyp($1,$3,OBJ_GOTOACT,GOTO_NOW);
		}
		| optional_ident SET IDENT '=' menustring {
                  $$=create_set_actiontyp($1,$3,$5,SET_DIRECT);
		}
		| optional_ident SET IDENT '=' IDENT {
                  $$=create_set_actiontyp($1,$3,$5,SET_INDIRECT);
		}
		| optional_ident SET IDENT '=' expressionstring {
		  char *s;
		  s=str_append3(strdup("set "),$3,strdup(" = "),$5);
                  $$=create_set_actiontyp($1,0,s,SET_EVALUATE);
		}
		;
		
optional_ident: IDENT ':' { $$=$1; }
		| {$$=(void *) 0; }
		;
		
arglist:	'(' arglistlist ')' { $$=$2; }
		;

arglistlist:	arglistitem {$$=(void *)list_append(0,(long)$1); }
		| arglistlist ',' arglistitem {
		  $$=(void *)list_append((long *)$1,(long)$3);
		}
		| arglistlist ',' {
		  $$=(void *)list_append((long *)$1,(long)0);
		}
		;


arglistitem:	expression {
		  $$=(void *)create_objekt(0,(void *)$1,OBJ_ARG_I);
		}
		| menustring {
		  $$=(void *)create_objekt(0,(void *)$1,OBJ_ARG_S);
		}
		;
		| IDENT {
		  $$=(void *)create_objekt(0,(void *)$1,OBJ_ARG_ID);
		}
		;

menustring:	STRING { $$=string_append($1,0); }
		| menustring STRING { $$=string_append($1,$2); }
		;


expressionstring: '(' expstring ')' { $$=str_append2(strdup("("),$2,strdup(")")); }
		;


expstring:	UINTEGER {
		  char s[40];
		  sprintf(s,"%ld",$1);
		  $$=strdup(s);
		}
		| IDENT {$$=str_append2(strdup(" "),$1,strdup(" "));}
		| expstring '&' expstring { $$=str_append2($1,strdup(" & "),$3); }
		| expstring '^' expstring { $$=str_append2($1,strdup(" ^ "),$3); }
		| expstring '|' expstring { $$=str_append2($1,strdup(" | "),$3); }
		| expstring SHL_OP expstring { $$=str_append2($1,strdup(" << "),$3); }
		| expstring SHR_OP expstring { $$=str_append2($1,strdup(" >> "),$3); }
		| expstring '+' expstring { $$=str_append2($1,strdup(" + "),$3); }
		| expstring '-' expstring { $$=str_append2($1,strdup(" - "),$3); }
		| expstring '*' expstring { $$=str_append2($1,strdup(" * "),$3); }
		| expstring '/' expstring { $$=str_append2($1,strdup(" / "),$3); }
		| expstring '%' expstring { $$=str_append2($1,strdup(" % "),$3); }
		| '(' expstring ')'  {
		  $$=str_append2(strdup("("),$2,strdup(")"));
		}
		| '-' expstring %prec UMINUS = {
		  $$ = string_append(strdup("-"),$2);
		}
		| '~' expstring %prec UCOMPL = {
		  $$ = string_append(strdup("~"),$2);
		}
                ;                                                                                     


expression: 	integer_expression
		;

integer_expression:
		UINTEGER
		| IDENT2 {
		  char *s = get_string_by_name($1);
		  if (s) $$=strtol(s,0,0);
		  else $$=0;
		  free($1);
		}
		| integer_expression '&' integer_expression   = {
		  $$ = $1 & $3;
		}
		| integer_expression '^' integer_expression   = {
		  $$ = $1 ^ $3;
		}
		| integer_expression '|' integer_expression   = {
		  $$ = $1 | $3;
		}
		| integer_expression SHL_OP integer_expression   = {
		  $$ = $1 << $3;
		}
		| integer_expression SHR_OP integer_expression   = {
		  $$ = $1 >> $3;
		}
		| integer_expression '+' integer_expression   = {
		  $$ = $1 + $3;
		}
		| integer_expression '-' integer_expression   = {
		  $$ = $1 - $3;
		}
		| integer_expression '*' integer_expression   = {
		  $$ = $1 * $3;
		}
		| integer_expression '/' integer_expression   = {
		  if($3 != 0) $$ = $1 / $3;
		  else {
		    if (FIRST_PASS) yyerror("division by zero !");
		    else $$=0;
		  }
		}
		| integer_expression '%' integer_expression   = {
		  if($3 != 0) $$ = $1 % $3;
		  else {
		    if (FIRST_PASS) yyerror("division by zero !");
		    else $$=0;
		  }
		}
		| '(' integer_expression ')' = { $$ = $2; }
		| '-' integer_expression %prec UMINUS = { $$ = -$2; }
		| '~' integer_expression %prec UCOMPL = { $$ = ~$2; }
                ;                                                                                     

%%

void yywarn(char* string, ...)
{
  va_list vars;
  va_start(vars, string);
  fprintf(stderr, "Warning: ");
  vfprintf(stderr, string, vars);
  fprintf(stderr, "\n");
  va_end(vars);
  warnings++;
}

void yyerror(char* string, ...)
{
  va_list vars;
  va_start(vars, string);
  fprintf(stderr, "Error: (file %s, line %.3d) ",include_fnames[include_stack_ptr], line_count);
  vfprintf(stderr, string, vars);
  fprintf(stderr, "\n");
  va_end(vars);
  errors++;
}

/*
 * open_file - opens the menu-file named *filename and returns
 *             a file-pointer. The error/warning-counters are reset to zero.
 */

FILE *open_file(char *filename)
{
  errors   = 0;                  /* Reset error counter */
  warnings = 0;                  /* Reset counter for warnings */

  include_fnames[0]=strdup(filename);
  return fopen(filename, "r"); /* Open menu-file */
}

/*
 * close_file - Close the menu file and issue a message about
 *              errors/warnings that occured.
 */
void close_file(FILE * file)
{
  if (fileno(file) != fileno(stdin)) fclose(file); /* Close the menu-file */

  if(errors)
    fprintf(stderr, "%d error(s) detected while parsing the menu-file\n",errors);
  if(warnings)
    fprintf(stderr, "%d warning(s) detected while parsing the menu-file\n",warnings);
#ifdef __linux__
  if (!(errors || warnings))
    fprintf(stderr, "menu-file Ok\n");
#endif
}


void *get_object_by_name(char *name, int kind)
{
  struct objectheader *p=objects;
  while (p) {
    if ( (kind <0) || (kind == p->kind) ) {
      if (!strcmp(p->olabel,name)) return p;
    }
    p=p->next;
  }
  return p;
}

char *get_string_by_name(char *name)
{
  struct objecttyp *p=get_object_by_name(name,  OBJ_VARSTRING);
  if (!p) return 0;
  return p->obj;
}


char *set_string_by_name(char *name, char *newvalue)
{
  struct genobjecttyp *p=get_object_by_name(name,  OBJ_VARSTRING);
  char *s;
  if (!p) return 0;
  s=strdup(newvalue);
  free(p->u.s);
  p->u.s=s;
  return s;
}


callbackfunction *get_callbackfunction_by_name(char *name)
{
  struct objecttyp *p=get_object_by_name(name,  OBJ_CALLBACK);
  return (void *)p->obj;  
}

int register_callbackfunction(callbackfunction *func, char *name)
{
  struct objecttyp *p=create_objekt(name, (void *)func, OBJ_CALLBACK);
  if (p) return 0;
  return 1;
}

char *get_startmenu_name()
{
  struct objecttyp *p=get_object_by_name("start",  OBJ_VARSTART);
  return p->obj;  
}


void *create_objekt(char *name, void *obj, int kind)
{
  struct objecttyp *p;
  
  if (name && ((p=get_object_by_name(name, -1)) != 0) ) {
    if (p->h.kind != kind) return 0;
    if (p->obj && (kind != OBJ_VARINTEGER)) free(p->obj);
  }
  else {
    p=malloc(sizeof(struct objecttyp));
    p->h.next=objects; /* chain into list */
    objects=(void *)p;
    p->h.kind=kind;
    p->h.olabel=name;
  }
  p->obj=obj;
  return (void *)p;
}


static long *list_append(long *list, long value)
{
  if (!list) {
    list=malloc(2* sizeof(long));
    list[0]=1;
    list[1]=value;
  }
  else {
    list=realloc(list,(list[0]+2)*sizeof(long));
    list[++list[0]]=value;
  }
  return list;
}

char *string_append(char *s1, char *s2)
{
  int size1,size2;
  int i;
  char *s;

  if (s2) s=s2;
  else s=s1;
  i=0;
  while (s[i]) {
    if (s[i] == '\\') {
      i++;
      switch (s[i]) {
        case 0: i--; break;
        case 'n': s[i]=10; break;
        case 'r': s[i]=13; break;
      }
    }
    *s=s[i];
    s++;
  }  
  *s=0;
  if (s2) {
    size1=strlen(s1);
    size2=strlen(s2)+1;
    s=malloc(size1+size2);
    memcpy(s,s1,size1);
    memcpy(s+size1,s2,size2);
    free(s1);
    free(s2);
    return s;
  }
  else return s1;
}


static char *str_append(char *s1, char *s2)
{
  int size1,size2;
  char *s;

  if (s2) {
    size1=strlen(s1);
    size2=strlen(s2)+1;
    s=malloc(size1+size2);
    memcpy(s,s1,size1);
    memcpy(s+size1,s2,size2);
    free(s1);
    free(s2);
    return s;
  }
  else return s1;
}


static char *str_append2(char *s1, char *s2, char *s3)
{
  char *s;
  s=str_append(s1,s2);
  return str_append(s,s3);
}


static char *str_append3(char *s1, char *s2, char *s3, char *s4)
{
  char *s;
  s=str_append(s1,s2);
  s=str_append(s,s3);
  return str_append(s,s4);
}



static void *create_generic_actiontyp(char *ident, void *obj, int kind, int flags)
{
  struct objecttyp *p=malloc(sizeof(struct objecttyp));
  p->h.next=0;
  p->h.kind=kind;
  p->h.flags=flags;
  p->h.olabel=ident;
  p->obj=obj;
  return (void *)p;
}


static void *create_set_actiontyp(char *ident, char *dest, char *src, int flags)
{
  struct set_actiontyp *p=malloc(sizeof(struct set_actiontyp));
  p->h.next=0;
  p->h.kind=OBJ_SETACT;
  p->h.flags=flags;
  p->h.olabel=ident;
  p->dest=dest;
  p->src=src;
  return (void *)p;
}


static void *create_show_actiontyp(char *ident, char *menuname, struct objecttyp *image, void **action)
{
  struct show_actiontyp *p=malloc(sizeof(struct show_actiontyp));
  p->h.next=0;
  p->h.kind=OBJ_SHOWACT;
  p->h.olabel=ident;
  p->menuname=menuname;
  p->image=image;
  p->action=(void *)action;
  return (void *)p;
}


static void *create_waitkey_actiontyp(char *ident, char *possible_strokes)
{
  struct waitkey_actiontyp *p=malloc(sizeof(struct waitkey_actiontyp));
  p->h.next=0;
  p->h.kind=OBJ_WAITKEYACT;
  p->h.olabel=ident;
  p->possible_strokes=possible_strokes;
  return (void *)p;
}

static void *create_call_actiontyp(char *ident,char *function,void *arg,
                                   int kind, int flags)
{
  struct call_actiontyp *p=malloc(sizeof(struct call_actiontyp));
  p->h.next=0;
  p->h.kind=kind;
  p->h.flags=flags;
  p->h.olabel=ident;
  p->function=function;
  p->arg=arg;
  return (void *)p;
}

static void *create_menu_body(char *image, void *menuaction)
{
  void **p=malloc(2*sizeof(void *));
  p[0]=image;
  p[1]=menuaction;
  return p;
}

static void *create_action_body(void *head, void *tail)
{
  void **p=(void **)head, **p_, *pp;
  int num=2;
  p_=p;
  if (tail) {
    while (*(p++)) num++;
  }
  p=malloc(num*sizeof(void *));
  pp=(void *)p;
  if (tail) {
    while (*p_) *(p++)=*(p_++);
    *(p++)=tail;
  }
  else *(p++)=head;
  *(p++)=0;
  if (tail) free(head);
  return pp;  
}


static void create_menu_obj(char *mlabel, struct genobjecttyp **geom, void *body)
{
  struct menutyp *m=malloc(sizeof(struct menutyp));
  int geoml=((long *)geom)[0];

  m->h.next=objects; /* chain into list */
  objects=(void *)m;

  m->h.kind=OBJ_MENU;
  m->h.olabel=mlabel;
  m->x=geom[1]->u.i;
  if (geoml <1) m->x=0;
  m->y=geom[2]->u.i;
  if (geoml <2) m->y=0;
  m->xsize=geom[3]->u.i;
  if (geoml <3) m->xsize=80-m->x;
  m->ysize=geom[4]->u.i;
  if (geoml <4) m->ysize=25-m->y;
  m->framepattern=0;
  if (geoml >4) m->framepattern=geom[5]->u.s;
#if 0
  /* NO, NO, mustn't free the the objects,
     because they are linked into the objekt list !!!!
   */
  {
    int i;
    for (i=1; i<=geoml; i++) free(geom+i);
  }
#endif
  m->frameattrib=-1;
  m->normattrib=-1;
  m->inversattrib=-1;
  m->highlightattrib=-1;

  if (geoml >5) m->frameattrib=geom[6]->u.i;
  if (geoml >6) m->normattrib=geom[7]->u.i;
  if (geoml >7) m->inversattrib=geom[8]->u.i;
  if (geoml >8) m->highlightattrib=geom[9]->u.i;
  m->userdata=0;

  free(geom);
  m->image=((void **)body)[0];
  m->action=(void **)((void **)body)[1];
#ifdef TESTPARSER
#ifdef __linux__
  printf("%s: menu ( %d,%d,%d,%d ) = {\n\"%s\"\n",
    m->h.olabel,
    m->x,m->y,m->xsize,m->ysize,
    (char *)((struct objecttyp *)m->image)->obj );
  if (m->action) {
    void **a=m->action;
    printf("  action {\n");
    while (a[0]) {
      switch (((struct objectheader *)a[0])->kind) {
        case OBJ_NONE: 
        case OBJ_MENU: break;
        case OBJ_SHOWACT: {
          printf("    show { %s }\n",((struct show_actiontyp *)a[0])->menuname);
          break;
        }
        case OBJ_WAITKEYACT: {
          printf("    waitkey { \"%s\" }\n",((struct waitkey_actiontyp *)a[0])->possible_strokes);
          break;
        }
      }
      a++;
    }
    printf("  }\n");
  }
  printf("}\n");
#endif
#endif
}
      
int single_parse(char *string_to_pars)
{
  int ret;
  errors   = 0;                  /* Reset error counter */
  warnings = 0;                  /* Reset counter for warnings */
  yyin=0;                        /* say: we have no input file */
  yy_vbuffer=string_to_pars;     /* this is the input to scan */
  line_count = 10001;            /* this to determin the source of error */
  ret = yyparse();
#ifdef __TURBOC__
  free_alloca();
#endif
  return (ret || errors);
}

int parse_menu(char *menu_name)
{
  FILE *volatile fd;
  int ret;
#if YYDEBUG != 0
  extern int yydebug;

  yydebug  = 1;
#endif

  /* Let's try menu_name if not null, and fail if not found */
  if(menu_name) {
    if((fd = open_file(menu_name)) == 0) {
      fprintf(stderr, "Cannot open  file %s, Aborting.\n",menu_name);
      return 1;
    }
  }
  else {
    fd = stdin;
  }

  yyin = fd;
  line_count = 1;
  ret = yyparse();
#ifdef __TURBOC__
  free_alloca();
#endif  
  close_file(fd);
  return (ret || errors);
}

#ifdef __linux__
int main(int argc, char** argv)
{
  if (!argv[1]) {
    fprintf(stderr, "need file to pars as first arg\n");
    exit(1);
  }
  parse_menu(argv[1]);
#if 0
  {
    char *s=get_string_by_name("second_pass");
    if (s) single_parse(s);
  }
#endif
  return 0;
}
#endif
