/*  >>> this is file parserd.h
        parser.c parserd.h should compile with GCC
        and with TURBO-C in large model
============================================================================
   LSETUP v1.0 (C) 1995 Hans Lermen (lermen@elserv.ffm.fgan.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


============================================================================
*/
#ifndef __PARSERD_H_
#define __PARSERD_H_

/* 
 * We need a lexical tie-in kludge for the second pass stuff
 * and we use yyin as switch:
 * if yyin !=0 we have an inputfile, so we are in first pass of parser run
 */
#define FIRST_PASS yyin 
                           

enum {
  OBJ_NONE,
  OBJ_MENU,
  OBJ_FUNCTION,
  OBJ_COLORS,
  OBJ_SHOWACT,
  OBJ_WAITKEYACT,
  OBJ_PRINTXYACT,
  OBJ_CALLACT,
  OBJ_CALLBACK,
  OBJ_GOTOACT,
  OBJ_SETACT,

  OBJ_VARIABLE=    0x100,
  OBJ_VARSTART=    0x101,
  OBJ_VARSTRING=   0x102,
  OBJ_VARINTEGER=  0x103,

  OBJ_ARG=         0x200,
  OBJ_ARG_I=       0x201,
  OBJ_ARG_S=       0x202,
  OBJ_ARG_ID=      0x203
};

enum {
  GOTO_NOW,
  GOTO_IF_TRUE,
  GOTO_IF_FALSE,
  GOTO_CALL_TRUE,
  GOTO_CALL_FALSE,
  GOTO_IF_ERROR=    0x100,

  SET_DIRECT,
  SET_INDIRECT,
  SET_EVALUATE
};


struct objectheader {
  struct objectheader *next;
  short int kind;
  short int flags;
  char *olabel;
};

struct objecttyp {
  struct objectheader h;
  void *obj;
};

struct genobjecttyp {
  struct objectheader h;
  union {
    void *obj;
    long i;
    unsigned long u;
    char *s;
  } u;
};

struct arglisttyp {
  long argc;
  struct genobjecttyp *argv[0];
};

struct menutyp {
  struct objectheader h;
  unsigned short int x,y,xsize,ysize;
  char *framepattern;
  int frameattrib,normattrib,inversattrib,highlightattrib;
  struct objecttyp *image;
  void **action;
  void* userdata;
};

struct show_actiontyp {
  struct objectheader h;
  char *menuname;
  struct objecttyp *image;
  struct objecttyp *action; /* object pointing to void **action */
};

struct waitkey_actiontyp {
  struct objectheader h;
  char *possible_strokes;
};

struct call_actiontyp {
  struct objectheader h;
  char *function;
  struct arglisttyp *arg;
};

struct set_actiontyp {
  struct objectheader h;
  char *dest;
  char *src;
};

typedef long callbackfunction(struct arglisttyp *);

#ifdef __PARSER_itself
  struct objectheader *objects=0;
#else
  extern struct objectheader *objects;
#endif

int parse_menu(char *menu_name);
int single_parse(char *string_to_pars);
void *get_object_by_name(char *name, int kind);
void *create_objekt(char *name, void *obj, int kind);
char *get_string_by_name(char *name);
char *set_string_by_name(char *name, char *newvalue);
char *get_startmenu_name();
callbackfunction *get_callbackfunction_by_name(char *name);
int register_callbackfunction(callbackfunction *func, char *name);

#endif /* #ifndef __PARSERD_H_ */
