/*  >>> this is file ULOADLIN.C
        compile it with TURBO-C in large model
============================================================================
   LOADLIN v1.6 (C) 1994,1996 Hans Lermen (lermen@elserv.ffm.fgan.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   As an exception to the above GNU General Public License (GPL):

   Usage of ULOADLIN.[CH] (not LOADLIN itself) is allowed within
   free usable executables even if the source for those programs
   is not provided, as long as all other terms of the GPL are respected.
   In all other cases usage of ULOADLIN.[CH] underlies the GPL.

----------------------------------------------------------------------------
   Comments and bug reports are welcome and may be sent to:
   E-Mail:    lermen@elserv.ffm.fgan.de
   SnailMail: Hans Lermen
              Am Muehlenweg 38
              D53424 REMAGEN-Unkelbach
              GERMANY

============================================================================ */

#if 1
  #include "_malloc_.h" /* malloc, realloc intercept, must be placed TOP */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include <io.h>
#include <fcntl.h>
#include <dos.h>
#include <sys/stat.h>

#define __ULOADLIN_itself
#include "uloadlin.h"

typedef struct {
  unsigned short opint20;
  unsigned short memend_frame;
  unsigned char  dos_reserved4;
  unsigned char  cpm_function_entry[0xa-0x5];
  unsigned long  int22_copy;
  unsigned long  int23_copy;
  unsigned long  int24_copy;
  unsigned short PID;
  unsigned char  file_handles[20];
  unsigned short envir_frame;
  unsigned long  system_stack;
  unsigned short max_open_files;
  unsigned long  file_handles_ptr;
  unsigned char  dos_reserved38[0x50-0x38];
  unsigned char  high_language_dos_call[0x53-0x50];
  unsigned char  dos_reserved53[0x5c-0x53];
  unsigned char  FCB1[0x6c-0x5c];
  unsigned char  FCB2[0x80-0x6c];
  unsigned char  DTA[0x100-0x80];
} dos_segprefix;


static char *err_tx[]={
    0,
    "cannot find LOADLIN.EXE",
    "not enough memory at top of 640 K lowmem, can\'t pass params to LOADLIN.EXE",
    "not enough memory starting at 0x90000, can\'t pass params to LOADLIN.EXE"
};

#define MAX_LOADLIN_ARGS 2
char *loadlin_argv[MAX_LOADLIN_ARGS+1]={0};


static void err_exit(char *tx)
{
	if (tx) {
		fprintf(stderr, "%s\n",tx);
		exit(1);
	}
	else exit(0);
}


char *get_loadlin_exe_path(char *pathprefix)
{
  int i,l,f;
  char *lname=loadlin_executable_path;
  struct stat sb;
  struct {
    long relfileoff;
    long flags;
    char level;
    char magic[8];
    char version[3];
  } suffix;

  if (pathprefix && pathprefix[0]) {
    strcpy(lname,pathprefix);
    l=strlen(lname);
    i=l;
    if (!stricmp(&lname[i-4],".EXE") || !stricmp(&lname[i-4],".COM")) {
      /* we got argv[0] as pathprefix */
      for (i=l; ((lname[i] !='\\') && (lname[i] !='/')) ; i--);
      i++;
    }
    else {
      /* we got a directory path as pathprefix */
      i=l;
      if ( (lname[i-1] != '\\') && (lname[i-1] != '/') ) lname[i++]='\\';
    }
  }
  else {
    lname[0]=0;
    i=0;
  }
  strcpy(lname+i,"LOADLIN.EXE");
  if (stat(lname, &sb)) return 0;
  f=open(lname,O_RDONLY);
  lseek(f, -(long)sizeof(suffix),SEEK_END);
  read(f,&suffix,sizeof(suffix));
  loadlin_version=0x105;
  if (!strncmp(suffix.magic,"Loadlin-",sizeof(suffix.magic))) {
    loadlin_version=(suffix.version[0]-'0' << 8);
    loadlin_version +=(suffix.version[2]-'0');
  }
  close(f);
  return lname;
}


int check_mem_for_loadlin()
{
  dos_segprefix *psp=MK_FP(_psp,0);
  int ret;

  if (loadlin_version >= 0x106) return 0;
  if (psp->memend_frame <= 0x9800) {
    /* not enough memory at top of 640 K lowmem, can\'t pass params to LOADLIN.EXE */
    return 2;
  }
  return 0;
}




int check_exec_loadlin(char *pathprefix)
{
  dos_segprefix *psp=MK_FP(_psp,0);
  unsigned segp;
  int ret;

  if (!get_loadlin_exe_path(pathprefix)) {
    /* we cannot find the loadlin executable */
    return 1;
  }
  ret=allocmem(0x1000,&segp);
  if (ret == -1) {
    freemem(segp);
    if (loadlin_version >= 0x106) return 0;
  }
  if (ret==-1 && segp < KERNEL_BOOTSECT_SEG) {
    if (psp->memend_frame <= KERNEL_SETUP_STOP) {
      /* not enough memory at top of 640 K lowmem, can\'t pass params to LOADLIN.EXE */
      return 2;
    }
  }
  else {
    /* not enough memory starting at 0x90000, can\'t pass params to LOADLIN.EXE */
    return 3;
  }
  return 0;
}

#define PARAM_BUF_SIZE (2*1024)
static unsigned char *param_buf=(void *)((long)KERNEL_SETUP_SEG << 16);
static unsigned char aux_param_buf[PARAM_BUF_SIZE];
static unsigned int aux_param_bytes;

static build_param_buf(char **argv)
{
  dos_segprefix *psp=MK_FP(_psp,0);
  static unsigned char *param_buf_;

  if (loadlin_version >= 0x106) param_buf=(void *)((long)(psp->memend_frame-(PARAM_BUF_SIZE >>4)) << 16);
  param_buf_=aux_param_buf;
  #if 0 /* not here, we are starting with the first param to loadlin */
    argv++;  /* skip argv[0] ( the program name ) */
  #endif
  while (*argv) param_buf_+=sprintf(param_buf_,"%s\n",*argv++);
  *param_buf_=0;
  aux_param_bytes=(param_buf_-aux_param_buf)+1;
}

static void move_param_buf_high()
{
  memcpy(param_buf,aux_param_buf,aux_param_bytes);
}


void clean_up_system()
{
	extern	int		_atexitcnt;    /* count of atexit function pointers */
	extern	atexit_t	_atexittbl[];  /* array of atexit function pointers */
	/* call all function that have been registered via atexit() */
	while (_atexitcnt--) (*_atexittbl[_atexitcnt])();
}

void exec_loadlin(char *pathprefix, char **argv)
{
  int i;
  char *lname=loadlin_executable_path;
  static char loadlin_argv1_buf[13];
  if (!use_loadlin) return;
  if ((i=check_exec_loadlin(pathprefix)) != 0) {
    err_exit(err_tx[i]);
  }

  build_param_buf(argv); /* this builds params in an aux-buffer */
  loadlin_argv[0]=lname;
  strcpy(loadlin_argv1_buf,"@@loadlinx@@");
  if (loadlin_version >= 0x106) sprintf(loadlin_argv1_buf+8,"%04x",(int)FP_SEG(param_buf));
  loadlin_argv[1]=loadlin_argv1_buf;
  loadlin_argv[2]=0;

	clean_up_system();
  move_param_buf_high(); /* this moves the aux-buffer to the final params place */
  if (execv(lname,loadlin_argv) == -1) err_exit("can't execute LOADLIN.EXE");
}

#if 0
/* testing it with an own main() */

int main(int argc, char **argv)
{
   use_loadlin=1;
   loadlin_argv[0]="\\loadlin\\1_3_77.bzi";
   loadlin_argv[1]="-t";
   check_exec_loadlin("\\suse\\setup");
   exec_loadlin("",argv);
}


#endif
