/*  >>> this is file ULOADLIN.H for ULOADLIN.C
        compile it with TURBO-C in large model
============================================================================
   LOADLIN v1.6 (C) 1994,1996 Hans Lermen (lermen@elserv.ffm.fgan.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   As an exception to the above GNU General Public License (GPL):

   Usage of ULOADLIN.[CH] (not LOADLIN itself) is allowed within
   free usable executables even if the source for those programs
   is not provided, as long as all other terms of the GPL are respected
   and the name of LOADLIN.EXE is not changed.
   In all other cases usage of ULOADLIN.[CH] underlies the GPL.

----------------------------------------------------------------------------
   Comments and bug reports are welcome and may be sent to:
   E-Mail:    lermen@elserv.ffm.fgan.de
   SnailMail: Hans Lermen
              Am Muehlenweg 38
              D53424 REMAGEN-Unkelbach
              GERMANY

============================================================================ */

#ifndef __ULOADLIN_H_
#define __ULOADLIN_H_

#define KERNEL_BOOTSECT_SEG 0x9000
#define KERNEL_SETUP_SEG    0x9020

   /* KERNEL_SETUP_STOP is the maximum as defined by loadlin.
      it actually depends on the number of setup-sectors as defined
      in bootsector byte 0x1F1 */
#define KERNEL_SETUP_STOP   0x9200


#ifdef __ULOADLIN_itself
  char use_loadlin=0;
  char loadlin_executable_path[80]="";
  int loadlin_version=0x105;
#else
  extern char use_loadlin;
  extern char loadlin_executable_path[80];
  extern int loadlin_version;
#endif

int check_exec_loadlin(char *pathprefix);


/* "exec_loadlin()" terminates the current process and executes LOADLIN.EXE
 * "pathprefix" can be the original argv[0] from main(int argc, char **argv),
 * it is used to get the place where LOADLIN.EXE can be found. But you
 * also may have set "pathprefix" to any directory or filename of an
 * executable (.EXE, .COM is checked ). The resulting filename of loadlin
 * will always end with LOADLIN.EXE, and ...
 *
 *   NOTE: You are NOT allowed to change the name of LOADLIN.EXE.
 *
 * "argv" is the parameter list for LOADLIN and starts with argv[0]
 * (be not confused, argv[0] is the first param, not the program nane)
 * For security reasons exec_loadlin() should be called from
 * within main() after as many as possible local resources have been freed.
 * A fixed area at 0x90000..0x91fff is used to pass parameters to LOADLIN,
 * because the 128 Byte limit for the DOS commandline may be too small,
 * this is not a conflict, because this area yet *must* be free in order
 * to load Linux (setup.S is fix located at 0x90000).
 * Also all atexit registered functions are called before the
 * actual execv().
 */
void exec_loadlin(char *pathprefix, char **argv);

/* This checks if loadlin can be executed savely. It returns with:
 *  0  =  Ok
 *  1  =  cannot find LOADLIN.EXE
 *  2  =  not enough memory at top of 640 K lowmem, can\'t pass params to LOADLIN.EXE",
 *  3  =  not enough memory starting at 0x90000, can\'t pass params to LOADLIN.EXE"
 */
int check_exec_loadlin(char *pathprefix);
int check_mem_for_loadlin();

/*  This checks if LOADLIN.EXE can be found and returns the
 *  full pathname. For behavior of pathprefix see exec_loadlin().
 */
char *get_loadlin_exe_path(char *pathprefix);


#endif  /* __ULOADLIN_H_ */
